"""
This test file tests the lib/smtpserver.py
"""
import email
from email.mime.image import MIMEImage
import binascii
from privacyidea.lib.queue import get_job_queue

from tests.queuemock import MockQueueTestCase
from .base import MyTestCase
from privacyidea.lib.error import ResourceNotFoundError
from privacyidea.lib.smtpserver import (get_smtpservers, add_smtpserver,
                                        delete_smtpserver, get_smtpserver,
                                        SMTPServer)
from . import smtpmock
from smtplib import SMTPException

PNG_IMG = 'iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAYAAACNMs+9AAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAAsTAA' \
          'ALEwEAmpwYAAAAB3RJTUUH5AEeDxMYtXhk0QAAAB1pVFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoIEdJTVBk' \
          'LmUHAAAAO0lEQVQY02P8////fwYiABMDkYAFXYCRkRGFD7MQq4n///9nQHcRCzaF6KbiVIjNf0R7hnyFuIKVaB' \
          'MB6yUTDUpeapUAAAAASUVORK5CYII='


class SMTPServerTestCase(MyTestCase):

    def test_01_create_smtpserver(self):
        r = add_smtpserver(identifier="myserver", server="1.2.3.4")
        self.assertTrue(r > 0)
        r = add_smtpserver(identifier="myserver1", server="5.4.3.2")
        self.assertTrue(r)
        r = add_smtpserver(identifier="myserver2", server="1.2.3.4")
        self.assertTrue(r)

        server_list = get_smtpservers()
        self.assertTrue(server_list)
        self.assertEqual(len(server_list), 3)
        server_list = get_smtpservers(identifier="myserver")
        self.assertEqual(len(server_list), 1)
        self.assertTrue(server_list[0].config.identifier, "myserver")
        self.assertTrue(server_list[0].config.port, 25)

        servers_by_ip = get_smtpservers(server='1.2.3.4')
        self.assertEqual(len(servers_by_ip), 2)
        for server in ["myserver", "myserver1", "myserver2"]:
            r = delete_smtpserver(server)
            self.assertTrue(r > 0)

        server_list = get_smtpservers()
        self.assertEqual(len(server_list), 0)

    @smtpmock.activate
    def test_02_send_email(self):
        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=False)
        self.assertTrue(r > 0)

        with self.assertRaises(ResourceNotFoundError):
            get_smtpserver(None)

        server = get_smtpserver("myserver")
        smtpmock.setdata(response={"recp@example.com": (200, "OK")},
                         support_tls=False)
        r = server.send_email(["recp@example.com"], "Hallo", "Body")
        self.assertEqual(r, True)

        smtpmock.setdata(response={"recp@example.com": (550,
                                                        "Message rejected")},
                         support_tls=False)
        r = server.send_email(["recp@example.com"], "Hallo", "Body")
        self.assertEqual(r, False)

        # Use TLS
        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=True)
        self.assertTrue(r > 0)
        server = get_smtpserver("myserver")
        smtpmock.setdata(response={"recp@example.com": (200, "OK")},
                         support_tls=True)
        r = server.send_email(["recp@example.com"], "Hallo", "Body")
        self.assertEqual(r, True)

        # If we configure TLS but the server does not support this, we raise
        # an error
        smtpmock.setdata(response={"recp@example.com": (200, "OK")},
                         support_tls=False)
        self.assertRaises(SMTPException, server.send_email,
                          ["recp@example.com"], "Hallo", "Body")

        delete_smtpserver("myserver")

    def test_03_updateserver(self):
        r = add_smtpserver(identifier="myserver", server="100.2.3.4")
        self.assertTrue(r > 0)
        server_list = get_smtpservers(identifier="myserver")
        self.assertTrue(server_list[0].config.server, "100.2.3.4")

    def test_04_missing_configuration(self):
        self.assertRaises(ResourceNotFoundError, get_smtpserver, "notExisting")

    @smtpmock.activate
    def test_05_test_email_smtp(self):
        smtpmock.setdata(response={"recp@example.com": (200, "OK")},
                         support_tls=True)
        identifier = "newConfig"
        server = "mailsever"
        port = 25
        username = "mailsender"
        password = "secret"
        sender = "mailsender@example.com"
        tls = True
        recipient = "user@example.com"

        s = dict(identifier=identifier, server=server, port=port,
                 username=username, password=password, sender=sender,
                 tls=tls)
        r = SMTPServer.test_email(s, recipient,
                                  "Test Email from privacyIDEA",
                                  "This is a test email from privacyIDEA. "
                                  "The configuration %s is working." % identifier)
        self.assertTrue(r)
        parsed_email = email.message_from_string(smtpmock.get_sent_message())
        self.assertEqual(parsed_email.get_content_type(), 'text/plain', parsed_email)
        self.assertEqual(parsed_email.get('To'), recipient, parsed_email)
        self.assertEqual(parsed_email.get('Subject'), "Test Email from privacyIDEA", parsed_email)

        # Now with an already prepared MIME email
        msg = MIMEImage(binascii.a2b_base64(PNG_IMG))
        r = SMTPServer.test_email(s, recipient, "Test Email with image",
                                  msg)
        self.assertTrue(r)
        parsed_email = email.message_from_string(smtpmock.get_sent_message())
        self.assertEqual(parsed_email.get_content_type(), 'image/png', parsed_email)
        self.assertEqual(parsed_email.get('To'), recipient, parsed_email)
        self.assertEqual(parsed_email.get('Subject'), "Test Email with image", parsed_email)
        # Check, that the mock SMTP server actually has NOT been configured as SMTP_SSL
        self.assertFalse(smtpmock.get_smtp_ssl())

    @smtpmock.activate
    def test_06_test_email_smtp_ssl(self):
        smtpmock.setdata(response={"recp@example.com": (200, "OK")},
                         support_tls=False)
        identifier = "newConfig"
        server = "smtps://mailserver"
        port = 25
        username = "mailsender"
        password = "secret"
        sender = "mailsender@example.com"
        tls = False
        recipient = "user@example.com"

        s = dict(identifier=identifier, server=server, port=port,
                 username=username, password=password, sender=sender,
                 tls=tls)
        r = SMTPServer.test_email(s, recipient,
                                  "Test Email from privacyIDEA",
                                  "This is a test email from privacyIDEA. "
                                  "The configuration %s is working." % identifier)
        self.assertTrue(r)
        parsed_email = email.message_from_string(smtpmock.get_sent_message())
        self.assertEqual(parsed_email.get_content_type(), 'text/plain', parsed_email)
        self.assertEqual(parsed_email.get('To'), recipient, parsed_email)
        self.assertEqual(parsed_email.get('Subject'), "Test Email from privacyIDEA", parsed_email)

        # Now with an already prepared MIME email
        msg = MIMEImage(binascii.a2b_base64(PNG_IMG))
        r = SMTPServer.test_email(s, recipient, "Test Email with image",
                                  msg)
        self.assertTrue(r)
        parsed_email = email.message_from_string(smtpmock.get_sent_message())
        self.assertEqual(parsed_email.get_content_type(), 'image/png', parsed_email)
        self.assertEqual(parsed_email.get('To'), recipient, parsed_email)
        self.assertEqual(parsed_email.get('Subject'), "Test Email with image", parsed_email)
        # Check, if the mock SMTP server actually has been configured as SMTP_SSL
        self.assertTrue(smtpmock.get_smtp_ssl())


class SMTPServerQueueTestCase(MockQueueTestCase):
    @smtpmock.activate
    def test_01_enqueue_email(self):
        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=False, enqueue_job=True)
        self.assertTrue(r > 0)

        server = get_smtpserver("myserver")
        smtpmock.setdata(response={"recp@example.com": (200, "OK")},
                         support_tls=False)
        r = server.send_email(["recp@example.com"], "Hallo", "Body")
        self.assertEqual(r, True)

        queue = get_job_queue()
        self.assertEqual(len(queue.enqueued_jobs), 1)
        job_name, args, kwargs = queue.enqueued_jobs[0]
        self.assertEqual(job_name, "smtpserver.send_email")
        self.assertEqual(args[1], ["recp@example.com"])
        self.assertEqual(args[2], "Hallo")
        self.assertEqual(args[3], "Body")

        # send_email returns True, even if the SMTP server will eventually reject the message
        smtpmock.setdata(response={"fail@example.com": (550,
                                                        "Message rejected")},
                         support_tls=False)
        r = server.send_email(["fail@example.com"], "Hallo", "Body")
        self.assertEqual(r, True)
        self.assertEqual(len(queue.enqueued_jobs), 2)
        job_name, args, kwargs = queue.enqueued_jobs[1]
        self.assertEqual(job_name, "smtpserver.send_email")
        self.assertEqual(args[1], ["fail@example.com"])
        self.assertEqual(args[2], "Hallo")
        self.assertEqual(args[3], "Body")

        delete_smtpserver("myserver")

    @smtpmock.activate
    def test_02_send_email_without_queue(self):
        # enqueue_job is False!
        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=False)
        self.assertTrue(r > 0)

        server = get_smtpserver("myserver")
        smtpmock.setdata(response={"recp@example.com": (200, "OK")},
                         support_tls=False)
        r = server.send_email(["recp@example.com"], "Hallo", "Body")
        self.assertEqual(r, True)

        smtpmock.setdata(response={"recp@example.com": (550,
                                                        "Message rejected")},
                         support_tls=False)
        r = server.send_email(["recp@example.com"], "Hallo", "Body")
        self.assertEqual(r, False)

        # Use TLS
        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=True)
        self.assertTrue(r > 0)
        server = get_smtpserver("myserver")
        smtpmock.setdata(response={"recp@example.com": (200, "OK")},
                         support_tls=True)
        r = server.send_email(["recp@example.com"], "Hallo", "Body")
        self.assertEqual(r, True)

        # If we configure TLS but the server does not support this, we raise
        # an error
        smtpmock.setdata(response={"recp@example.com": (200, "OK")},
                         support_tls=False)
        self.assertRaises(SMTPException, server.send_email,
                          ["recp@example.com"], "Hallo", "Body")

        # Assert that no
        queue = get_job_queue()
        self.assertEqual(queue.enqueued_jobs, [])

        delete_smtpserver("myserver")




#    2019-02-12 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
#
#    This program is free software: you can redistribute it and/or
#    modify it under the terms of the GNU Affero General Public
#    License, version 3, as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the
#               GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

__doc__ = """This is the provider class that communicates with Googles
Firebase Cloud Messaging Service.
This provider is used for the push token and can be used for SMS tokens.
"""

from privacyidea.lib.smsprovider.SMSProvider import (ISMSProvider)
from privacyidea.lib.error import ConfigAdminError
from privacyidea.lib.framework import get_app_local_store
from privacyidea.lib import _
import logging
from google.oauth2 import service_account
from google.auth.transport.requests import AuthorizedSession
import json
import time

FIREBASE_URL_SEND = 'https://fcm.googleapis.com/v1/projects/{0!s}/messages:send'
SCOPES = ['https://www.googleapis.com/auth/cloud-platform',
          'https://www.googleapis.com/auth/datastore',
          'https://www.googleapis.com/auth/devstorage.read_write',
          'https://www.googleapis.com/auth/firebase',
          'https://www.googleapis.com/auth/identitytoolkit',
          'https://www.googleapis.com/auth/userinfo.email']

log = logging.getLogger(__name__)


def get_firebase_access_token(config_file_name):
    """
    This returns the access token for a given JSON config file name

    :param config_file_name: The json file with the Service account credentials
    :type config_file_name: str
    :return: Firebase credentials
    :rtype: google.oauth2.service_account.Credentials
    """
    fbt = "firebase_token"
    app_store = get_app_local_store()

    if fbt not in app_store or not isinstance(app_store[fbt], dict):
        # initialize the firebase_token in the app_store as dict
        app_store[fbt] = {}

    if not isinstance(app_store[fbt].get(config_file_name), service_account.Credentials) or \
            app_store[fbt].get(config_file_name).expired:
        # If the type of the config is not of class Credentials or if the token
        # has expired we get new scoped access token credentials
        credentials = service_account.Credentials.from_service_account_file(config_file_name,
                                                                            scopes=SCOPES)

        log.debug("Fetching a new access_token for {!r} from firebase...".format(config_file_name))
        # We do not use a lock here: The worst that could happen is that two threads
        # fetch new auth tokens concurrently. In this case, one of them wins and
        # is written to the dictionary.
        app_store[fbt][config_file_name] = credentials
        readable_time = credentials.expiry.isoformat() if credentials.expiry else 'Never'
        log.debug("Setting the expiration for {!r} of the new access_token "
                  "to {!s}.".format(config_file_name, readable_time))

    return app_store[fbt][config_file_name]


class FIREBASE_CONFIG:
    REGISTRATION_URL = "registration URL"
    TTL = "time to live"
    JSON_CONFIG = "JSON config file"
    HTTPS_PROXY = "httpsproxy"


class FirebaseProvider(ISMSProvider):

    def __init__(self, db_smsprovider_object=None, smsgateway=None):
        ISMSProvider.__init__(self, db_smsprovider_object, smsgateway)
        self.access_token_info = None
        self.access_token_expires_at = 0

    def submit_message(self, firebase_token, data):
        """
        send a message to a registered Firebase client
        This can be a simple OTP value or a cryptographic challenge response.

        :param firebase_token: The firebase token of the smartphone
        :type firebase_token: str
        :param data: the data dictionary part of the message to submit to the phone
        :type data: dict
        :return: bool
        """
        res = False

        credentials = get_firebase_access_token(self.smsgateway.option_dict.get(
            FIREBASE_CONFIG.JSON_CONFIG))

        authed_session = AuthorizedSession(credentials)

        headers = {
            'Content-Type': 'application/json; UTF-8',
        }
        fcm_message = {
            "message": {
                        "data": data,
                        "token": firebase_token,
                        "notification": {
                            "title": data.get("title"),
                            "body": data.get("question")
                        },
                        "android": {
                                    "priority": "HIGH",
                                    "ttl": "120s",
                                    "fcm_options": {"analytics_label": "AndroidPushToken"}
                                   },
                        "apns": {
                                 "headers": {
                                             "apns-priority": "10",
                                             "apns-push-type": "alert",
                                             "apns-collapse-id": "privacyidea.pushtoken",
                                             "apns-expiration": str(int(time.time()) + 120)
                                            },
                                 "payload": {
                                             "aps": {
                                                     "alert": {
                                                               "title": data.get("title"),
                                                               "body": data.get("question"),
                                                              },
                                                     "sound": "default",
                                                     "category": "PUSH_AUTHENTICATION"
                                                    },
                                            },
                                 "fcm_options": {"analytics_label": "iOSPushToken"}
                                }
                       }
            }

        proxies = {}
        if self.smsgateway.option_dict.get(FIREBASE_CONFIG.HTTPS_PROXY):
            proxies["https"] = self.smsgateway.option_dict.get(FIREBASE_CONFIG.HTTPS_PROXY)
        a = self.smsgateway.option_dict.get(FIREBASE_CONFIG.JSON_CONFIG)
        with open(a) as f:
            server_config = json.load(f)
        url = FIREBASE_URL_SEND.format(server_config["project_id"])
        resp = authed_session.post(url, data=json.dumps(fcm_message), headers=headers, proxies=proxies)

        if resp.status_code == 200:
            log.debug("Message sent successfully to Firebase service.")
            res = True
        else:
            log.warning("Failed to send message to firebase service: {0!s}".format(resp.text))

        return res

    def check_configuration(self):
        """
        This method checks the sanity of the configuration of this provider.
        If there is a configuration error, than an exception is raised.
        :return:
        """
        json_file = self.smsgateway.option_dict.get(FIREBASE_CONFIG.JSON_CONFIG)
        server_config = None
        with open(json_file) as f:
            server_config = json.load(f)
        if server_config:
            if server_config.get("type") != "service_account":
                raise ConfigAdminError(description="The JSON file is not a valid firebase credentials file.")

        else:
            raise ConfigAdminError(description="Please check your configuration. Can not load JSON file.")


    @classmethod
    def parameters(cls):
        """
        Return a dictionary, that describes the parameters and options for the
        Firebase Provider
        Parameters are required keys to values.

        :return: dict
        """
        params = {"options_allowed": False,
                  "headers_allowed": False,
                  "parameters": {
                      FIREBASE_CONFIG.JSON_CONFIG: {
                          "required": True,
                          "description": _("The filename of the JSON config file, that allows privacyIDEA to talk"
                                           " to the Firebase REST API.")
                      },
                      FIREBASE_CONFIG.HTTPS_PROXY: {
                          "required": False,
                          "description": _("Proxy setting for HTTPS connections to googleapis.com.")
                      }
                  }
                  }
        return params



# SPDX-FileCopyrightText: (C) 2024 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-FileCopyrightText: (C) 2016 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.
__doc__ = """
This script searches the OTP value in a given list of tokens.
The searched tokens are determined by tokentype, substring of the serial,
assigned status...

The serial number of the token is returned.

You can call the script like this:

    privacyidea-get-serial byotp --otp <otp> --type <type> --serial <serial>
        --unassigned --assigned --window <window>

"""
__version__ = "0.1"

import click
from flask.cli import with_appcontext, ScriptInfo

from privacyidea.lib.token import get_tokens, get_serial_by_otp
from privacyidea.lib.utils import get_version_number
from privacyidea.cli import create_silent_app

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-a', '--assigned', is_flag=True,
              help='If set, search only assigned tokens')
@click.option('-u', '--unassigned', is_flag=True,
              help='If set, search only unassigned tokens')
@click.option('-w', '--window', type=int, default=10, show_default=True,
              help="The OTP window for calculating OTP values. Default=10")
@click.option('-s', '--serial', type=str, default="",
              help="A part of the token serial number")
@click.option('-t', '--type', 'tokentype', type=str, default=None,
              help="The tokentype like hotp, totp, ...")
@click.argument('otp')
@with_appcontext
def byotp(otp, tokentype, serial, window, unassigned, assigned):
    """
    This searches the list of the specified tokens for the given OTP value.
    """
    if not assigned and not unassigned:
        assigned = None
    count = get_tokens(tokentype=tokentype, serial_wildcard="*{0!s}*".format(
            serial), assigned=assigned, count=True)
    click.echo("Searching in {0!s} tokens.".format(count))

    tokenobj_list = get_tokens(tokentype=tokentype,
                               serial_wildcard="*{0!s}*".format(serial),
                               assigned=assigned)
    serial = get_serial_by_otp(tokenobj_list, otp=otp, window=window)
    if serial:
        click.echo("Found the token with serial {0!s}".format(serial))
    else:
        click.echo("No token found.")


def byotp_call():
    """Call the actual function with an initialized app"""
    click.echo(r"""
                 _                    _______  _______
       ___  ____(_)  _____ _______ __/  _/ _ \/ __/ _ |
      / _ \/ __/ / |/ / _ `/ __/ // // // // / _// __ |
     / .__/_/ /_/|___/\_,_/\__/\_, /___/____/___/_/ |_|  Get Serial
    /_/                       /___/
    {0!s:>51}
        """.format('v{0!s}'.format(get_version_number())))

    # Add the ScriptInfo object to create the Flask-App when necessary
    s = ScriptInfo(create_app=create_silent_app)
    byotp(obj=s)


if __name__ == '__main__':
    byotp_call()



#  2019-09-14 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Initial writup
#
# License:  AGPLv3
# (c) 2019. Cornelius Kölbel
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """This is the event handler module that can mangle the JSON response.
We can add or delete key or even subtrees in the JSON response of a request.

The key is identified by a JSON Pointer
(see https://tools.ietf.org/html/rfc6901)
"""
from privacyidea.lib.eventhandler.base import BaseEventHandler
from privacyidea.lib.utils import is_true
from privacyidea.lib import _
import json
import logging


log = logging.getLogger(__name__)


class ACTION_TYPE(object):
    """
    Allowed actions
    """
    SET = "set"
    DELETE = "delete"


class ResponseManglerEventHandler(BaseEventHandler):
    """
    An Eventhandler needs to return a list of actions, which it can handle.

    It also returns a list of allowed action and conditions

    It returns an identifier, which can be used in the eventhandlig definitions
    """

    identifier = "ResponseMangler"
    description = "This event handler can mangle the JSON response."

    @property
    def allowed_positions(cls):
        """
        This returns the allowed positions of the event handler definition.
        The ResponseMangler can only be located at the "post" position

        :return: list of allowed positions
        """
        return ["post"]

    @property
    def actions(cls):
        """
        This method returns a dictionary of allowed actions and possible
        options in this handler module.

        :return: dict with actions
        """
        actions = {ACTION_TYPE.DELETE:
                       {"JSON pointer":
                            {"type": "str",
                             "required": True,
                             "description": _("The JSON pointer (key) that should be deleted. Please "
                                              "specify in the format '/detail/message'.")}
                        },
                   ACTION_TYPE.SET:
                       {"JSON pointer":
                            {"type": "str",
                             "required": True,
                             "description": _("The JSON pointer (key) that should be set. "
                                              "Please specify in the format '/detail/message'.")
                             },
                        "type":
                            {"type": "str",
                             "required": True,
                             "description": _("The type of the value."),
                             "value": ["string", "integer", "bool"]
                            },
                        "value":
                            {"type": "str",
                             "required": True,
                             "description": _("The value of the JSON key that should be set.")
                            }
                        }
                   }
        return actions

    def do(self, action, options=None):
        """
        This method executes the defined action in the given event.

        :param action:
        :param options: Contains the flask parameters g, request, response
            and the handler_def configuration
        :type options: dict
        :return:
        """
        ret = True
        response = options.get("response")
        try:
            content = self._get_response_content(response)
        except Exception:
            # The body could not be decoded to JSON. Actually this would be
            # a BadRequest, but we refrain from importing from werkzeug, here.
            log.info("The body could not be decoded to JSON.")
            # Early exist
            return ret
        handler_def = options.get("handler_def")
        handler_options = handler_def.get("options", {})

        json_pointer = handler_options.get("JSON pointer", "")
        value = handler_options.get("value")
        type = handler_options.get("type")
        comp = json_pointer.split("/")
        comp = [x for x in comp if x]
        if action.lower() == ACTION_TYPE.DELETE:
            try:
                if len(comp) == 1:
                    del(content[comp[0]])
                elif len(comp) == 2:
                    del(content[comp[0]][comp[1]])
                elif len(comp) == 3:
                    del (content[comp[0]][comp[1]][comp[2]])
                else:
                    log.warning("JSON pointer length of {0!s} not supported.".format(len(comp)))
                options.get("response").data = json.dumps(content)
            except KeyError:
                log.warning("Can not delete response JSON Pointer {0!s}.".format(json_pointer))
        elif action.lower() == ACTION_TYPE.SET:
            try:
                if type == "integer":
                    value = int(value)
                elif type == "bool":
                    value = is_true(value)
            except ValueError:
                log.warning("Failed to convert value")

            if len(comp) == 1:
                content[comp[0]] = value
            elif len(comp) == 2:
                content[comp[0]] = content.get(comp[0]) or {}
                content[comp[0]][comp[1]] = value
            elif len(comp) == 3:
                content[comp[0]] = content.get(comp[0]) or {}
                content[comp[0]][comp[1]] = content[comp[0]].get(comp[1]) or {}
                content[comp[0]][comp[1]][comp[2]] = value
            else:
                log.warning("JSON pointer of length {0!s} not supported.".format(len(comp)))
            options.get("response").data = json.dumps(content)

        return ret



#  2018-10-31 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Let the client choose to get a HTTP 500 Error code if
#             SMS fails.
#  2018-02-16 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add dynamic email address. Dynamically read from user source.
#  2018-01-21 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add email templates
#  2015-12-29 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Use privacyidea.lib.smtpserver instead of smtplib
#  2015-10-12 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add test config function
#  2015-04-14 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Adapt code to work with privacyIDEA 2 (Flask)
#
#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  LSE
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This is the implementation of an Email-Token, that sends OTP
values via SMTP.

The following config entries are used:

 * email.validtime
 * email.identifier

The identifier points to a system wide SMTP server configuration.
See :ref:`rest_smtpserver`.

The system-wide SMTP server configuration was introduced in version 2.10.
In privacyIDEA up to version 2.9 the following config entries were used:

 * email.validtime
 * email.mailserver
 * email.port
 * email.username
 * email.password
 * email.mailfrom
 * email.subject
 * email.tls

policy: action: emailtext

The code is tested in tests/test_lib_tokens_email
"""

import logging
import traceback
import datetime
from privacyidea.lib.tokens.smstoken import HotpTokenClass
from privacyidea.lib.tokenclass import CHALLENGE_SESSION, AUTHENTICATIONMODE
from privacyidea.lib.config import get_from_config, get_email_validators
from privacyidea.api.lib.utils import getParam
from privacyidea.lib.utils import is_true, create_tag_dict
from privacyidea.lib.policy import (SCOPE, ACTION, GROUP, comma_escape_text,
                                    get_action_values_from_options)
from privacyidea.lib.policy import Match
from privacyidea.lib.error import ValidateError
from privacyidea.lib.log import log_with
from privacyidea.lib import _
from privacyidea.models import Challenge
from privacyidea.lib.decorators import check_token_locked
from privacyidea.lib.smtpserver import send_email_data, send_email_identifier
from privacyidea.lib.crypto import safe_compare


log = logging.getLogger(__name__)
TEST_SUCCESSFUL = "Successfully sent email. Please check your inbox."


class EMAILACTION(object):
    EMAILTEXT = "emailtext"
    EMAILSUBJECT = "emailsubject"
    EMAILAUTO = "emailautosend"


class EmailTokenClass(HotpTokenClass):
    """
    Implementation of the EMail Token Class, that sends OTP values via SMTP.
    (Similar to SMSTokenClass)
    """

    EMAIL_ADDRESS_KEY = "email"
    # The HOTP token provides means to verify the enrollment
    can_verify_enrollment = True
    mode = [AUTHENTICATIONMODE.CHALLENGE]

    def __init__(self, aToken):
        HotpTokenClass.__init__(self, aToken)
        self.set_type("email")
        # we support various hashlib methods, but only on create
        # which is effectively set in the update
        self.hashlibStr = get_from_config("hotp.hashlib", "sha1")

    @property
    def _email_address(self):
        if is_true(self.get_tokeninfo("dynamic_email")):
            email = self.user.info.get(self.EMAIL_ADDRESS_KEY)
            if isinstance(email, list) and email:
                # If there is a non-empty list, we use the first entry
                email = email[0]
        else:
            email = self.get_tokeninfo(self.EMAIL_ADDRESS_KEY)
        if not email:  # pragma: no cover
            log.warning("Token {0!s} does not have an email address!".format(self.token.serial))
        return email

    @_email_address.setter
    def _email_address(self, value):
        self.add_tokeninfo(self.EMAIL_ADDRESS_KEY, value)

    @staticmethod
    def get_class_type():
        """
        return the generic token class identifier
        """
        return "email"

    @staticmethod
    def get_class_prefix():
        return "PIEM"

    @staticmethod
    def get_class_info(key=None, ret='all'):
        """
        returns all or a subtree of the token definition

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype: dict
        """
        res = {'type': 'email',
               'title': _('EMail Token'),
               'description':
                   _('EMail: Send a One Time Password to the users email '
                     'address.'),
               'user': ['enroll'],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': ["admin", "user"],
               'policy': {SCOPE.AUTH: {
                   EMAILACTION.EMAILTEXT: {
                       'type': 'str',
                       'desc': _('The text that will be sent via EMail for '
                                 'an EMail-token. Several tags like {otp} and '
                                 '{serial} can be used as parameters. You may '
                                 'also specify a filename as email template '
                                 'starting with "file:".')
                               + " " + comma_escape_text
                   },
                   EMAILACTION.EMAILSUBJECT: {
                       'type': 'str',
                       'desc': _('The subject of the EMail for '
                                 'an EMail token. Use tags like {otp} and {serial} '
                                 'as parameters.')},
                   EMAILACTION.EMAILAUTO: {
                       'type': 'bool',
                       'desc': _('If set, a new EMail OTP will be sent '
                                 'after successful authentication with '
                                 'one EMail OTP.')},
                   ACTION.CHALLENGETEXT: {
                       'type': 'str',
                       'desc': _('Use an alternative challenge text for telling the '
                                 'user to enter the code from the e-mail. You can also '
                                 'use tags for automated replacement. Check out the documentation '
                                 'for more details.')
                               + " " + comma_escape_text
                   },
               },
                   SCOPE.ENROLL: {
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of email tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of active email tokens assigned."),
                           'group': GROUP.TOKEN
                       }
                   }
               }
               }

        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res

        return ret

    @log_with(log)
    def update(self, param, reset_failcount=True):
        """
        update - process initialization parameters

        :param param: dict of initialization parameters
        :type param: dict

        :return: nothing

        """
        verify = getParam(param, "verify", optional=True)
        if not verify:
            if getParam(param, "dynamic_email", optional=True):
                self.add_tokeninfo("dynamic_email", True)
            else:
                # specific - e-mail
                self._email_address = getParam(param,
                                               self.EMAIL_ADDRESS_KEY,
                                               optional=False)

            # in case of the e-mail token, only the server must know the otpkey
            # thus if none is provided, we let create one (in the TokenClass)
            if 'genkey' not in param and 'otpkey' not in param:
                param['genkey'] = 1

        HotpTokenClass.update(self, param, reset_failcount)
        return

    @log_with(log)
    def is_challenge_request(self, passw, user=None, options=None):
        """
        check, if the request would start a challenge

        We need to define the function again, to get rid of the
        is_challenge_request-decorator of the HOTP-Token

        :param passw: password, which might be pin or pin+otp
        :param options: dictionary of additional request parameters

        :return: returns true or false
        """
        return self.check_pin(passw, user=user, options=options)

    @log_with(log)
    def create_challenge(self, transactionid=None, options=None):
        """
        create a challenge, which is submitted to the user

        :param transactionid: the id of this challenge
        :param options: the request context parameters / data
            You can pass ``exception=1`` to raise an exception, if
            the Email could not be sent.
        :return: tuple
            of (success, message, transactionid, attributes)

                * success: if submit was successful
                * message: the text submitted to the user
                * transactionid: the given or generated transactionid
                * reply_dict: additional dictionary, which is added to the response
        :rtype: tuple(bool, str, str, dict)
        """
        success = False
        options = options or {}
        return_message = get_action_values_from_options(SCOPE.AUTH,
                                                        "{0!s}_{1!s}".format(self.get_class_type(),
                                                                             ACTION.CHALLENGETEXT),
                                                        options) or _("Enter the OTP from the Email")

        return_message = return_message.replace(r'\,', ',')

        reply_dict = {'attributes': {'state': transactionid}}
        validity = int(get_from_config("email.validtime", 120))

        if self.is_active() is True:
            counter = self.get_otp_count()
            log.debug("counter={0!r}".format(counter))

            # At this point we must not bail out in case of a
            # Gateway error, since checkPIN is successful. A bailout
            # would cancel the checking of the other tokens
            try:
                data = None
                if options.get("session") != CHALLENGE_SESSION.ENROLLMENT:
                    # Only if this is NOT a multichallenge enrollment, we try to send the email
                    self.inc_otp_counter(counter, reset=False)
                    message_template, mimetype = self._get_email_text_or_subject(options)
                    subject_template, _n = self._get_email_text_or_subject(options,
                                                                           EMAILACTION.EMAILSUBJECT,
                                                                           "Your OTP")
                    success, sent_message = self._compose_email(
                        options=options,
                        message=message_template,
                        subject=subject_template,
                        mimetype=mimetype)

                # Create the challenge in the database
                if is_true(get_from_config("email.concurrent_challenges")):
                    data = self.get_otp()[2]
                db_challenge = Challenge(self.token.serial,
                                         transaction_id=transactionid,
                                         challenge=options.get("challenge"),
                                         data=data,
                                         session=options.get("session"),
                                         validitytime=validity)
                db_challenge.save()
                transactionid = transactionid or db_challenge.transaction_id

            except Exception as e:
                info = _("The PIN was correct, but the "
                         "EMail could not be sent!")
                log.warning(info + " ({0!r})".format(e))
                log.debug("{0!s}".format(traceback.format_exc()))
                return_message = info
                if is_true(options.get("exception")):
                    raise Exception(info)

        expiry_date = datetime.datetime.now() + \
                      datetime.timedelta(seconds=validity)
        reply_dict['attributes']['valid_until'] = "{0!s}".format(expiry_date)

        return success, return_message, transactionid, reply_dict

    @log_with(log)
    @check_token_locked
    def check_otp(self, anOtpVal, counter=None, window=None, options=None):
        """
        check the otpval of a token against a given counter
        and the window

        :param anOtpVal: the to be verified passw/pin
        :type anOtpVal: string

        :return: counter if found, -1 if not found
        :rtype: int
        """
        options = options or {}
        ret = HotpTokenClass.check_otp(self, anOtpVal, counter, window, options)

        if ret < 0 and is_true(get_from_config("email.concurrent_challenges")):
            if safe_compare(options.get("data"), anOtpVal):
                # We authenticate from the saved challenge
                ret = 1
        if ret >= 0 and self._get_auto_email(options):
            message, mimetype = self._get_email_text_or_subject(options)
            subject, _ = self._get_email_text_or_subject(options,
                                                         action=EMAILACTION.EMAILSUBJECT,
                                                         default="Your OTP")
            self.inc_otp_counter(ret, reset=False)
            success, message = self._compose_email(options=options,
                                                   message=message,
                                                   subject=subject,
                                                   mimetype=mimetype)
            log.debug("AutoEmail: send new SMS: {0!s}".format(success))
            log.debug("AutoEmail: {0!r}".format(message))
        return ret

    @staticmethod
    def _get_email_text_or_subject(options,
                                   action=EMAILACTION.EMAILTEXT,
                                   default="<otp>"):
        """
        This returns the EMAILTEXT or EMAILSUBJECT from the policy
        "emailtext" or "emailsubject

        :param options: contains user and g object.
        :type options: dict
        :param action: The action - either emailtext or emailsubject
        :param default: If no policy can be found, this is the default text
        :return: Message template, MIME type (one of "plain", "html")
        :rtype: (basestring, basestring)
        """
        message = default
        mimetype = "plain"
        g = options.get("g")
        user_object = options.get("user")
        if g:
            messages = Match.user(g, scope=SCOPE.AUTH, action=action, user_object=user_object if user_object else None)\
                .action_values(unique=True, allow_white_space_in_action=True)
            if len(messages) == 1:
                message = list(messages)[0]

        message = message.replace(r'\,', ',')

        if message.startswith("file:"):
            # We read the template from the file.
            try:
                with open(message[5:], "r") as f:
                    message = f.read()
                    mimetype = "html"
            except Exception as e:  # pragma: no cover
                message = default
                log.warning("Failed to read email template: {0!r}".format(e))
                log.debug("{0!s}".format(traceback.format_exc()))

        return message, mimetype

    @staticmethod
    def _get_auto_email(options):
        """
        This returns the AUTOEMAIL setting.

        :param options: contains user and g object.
        :type options: dict
        :return: True if an SMS should be sent automatically
        :rtype: bool
        """
        autosms = False
        g = options.get("g")
        user_object = options.get("user")
        if g:
            autoemailpol = Match.user(g, scope=SCOPE.AUTH, action=EMAILACTION.EMAILAUTO, user_object=user_object).policies()
            autosms = len(autoemailpol) >= 1

        return autosms

    @log_with(log)
    def _compose_email(self, message="<otp>", subject="Your OTP", mimetype="plain", options=None):
        """
        send email

        :param message: the email submit message - could contain placeholders
            like <otp> or <serial>
        :type message: string
        :param mimetype: the message MIME type - one of "plain", "html"
        :type mimetype: basestring

        :return: submitted message
        :rtype: string
        """
        ret = None

        recipient = self._email_address
        otp = self.get_otp()[2]
        serial = self.get_serial()
        options = options or {}
        challenge = options.get("challenge")

        message = message.replace("<otp>", otp)
        message = message.replace("<serial>", serial)

        tags = create_tag_dict(serial=serial,
                               tokenowner=self.user,
                               tokentype=self.get_tokentype(),
                               recipient={"givenname": self.user.info.get("givenname") if self.user else "",
                                          "surname": self.user.info.get("surname") if self.user else ""},
                               escape_html=mimetype.lower() == "html",
                               challenge=challenge)

        message = message.format(otp=otp, **tags)

        subject = subject.replace("<otp>", otp)
        subject = subject.replace("<serial>", serial)

        subject = subject.format(otp=otp, **tags)

        log.debug("sending Email to {0!r}".format(recipient))

        # The token specific identifier has priority over the system-wide identifier
        identifier = self.get_tokeninfo("email.identifier") or get_from_config("email.identifier")
        if identifier:
            # New way to send email
            ret = send_email_identifier(identifier, recipient, subject, message,
                                        mimetype=mimetype)
        else:
            # old way to send email / DEPRECATED
            mailserver = get_from_config("email.mailserver", "localhost")
            port = int(get_from_config("email.port", 25))
            username = get_from_config("email.username")
            password = get_from_config("email.password")
            mail_from = get_from_config("email.mailfrom", "privacyidea@localhost")
            email_tls = get_from_config("email.tls", default=False,
                                        return_bool=True)
            ret = send_email_data(mailserver, subject, message, mail_from,
                                  recipient, username, password, port,
                                  email_tls)
        return ret, message

    @classmethod
    def test_config(cls, params=None):
        mailserver = getParam(params, "email.mailserver", optional=False)
        subject = "Your TEST OTP"
        message = "This is a test."
        mail_from = getParam(params, "email.mailfrom", optional=False)
        recipient = getParam(params, "email.recipient", optional=False)
        password = getParam(params, "email.password")
        username = getParam(params, "email.username")
        port = getParam(params, "email.port", default=25)
        email_tls = getParam(params, "email.tls", default=False)
        r = send_email_data(mailserver, subject, message, mail_from,
                            recipient, username=username,
                            password=password, port=port, email_tls=email_tls)

        description = "Could not send email."
        if r:
            description = TEST_SUCCESSFUL

        return r, description

    def prepare_verify_enrollment(self, options=None):
        """
        This is called, if the token should be enrolled in a way, that the user
        needs to provide a proof, that the server can verify, that the token
        was successfully enrolled.
        The email token needs to send an email with OTP.

        The returned dictionary is added to the response in "detail" -> "verify".

        :return: A dictionary with information that is needed to trigger the verification.
        """
        # TODO: how should we handle errors when sending the mail?
        #  Currently we just return the message without checking the result
        _res, msg, _tid, _rdict = self.create_challenge(options=options)
        return {"message": msg}

    @classmethod
    def enroll_via_validate(cls, g, content, user_obj, message=None):
        """
        This class method is used in the policy ENROLL_VIA_MULTICHALLENGE.
        It enrolls a new token of this type and returns the necessary information
        to the client by modifying the content.

        :param g: context object
        :param content: The content of a response
        :param user_obj: A user object
        :param message: An alternative message displayed to the user during enrollment
        :return: None, the content is modified
        """
        from privacyidea.lib.token import init_token
        from privacyidea.lib.tokenclass import CLIENTMODE
        token_obj = init_token({"type": cls.get_class_type(),
                                "dynamic_email": 1}, user=user_obj)
        content.get("result")["value"] = False
        content.get("result")["authentication"] = "CHALLENGE"

        detail = content.setdefault("detail", {})
        # Create a challenge!
        options = {"session": CHALLENGE_SESSION.ENROLLMENT, "g": g, "user": user_obj}
        c = token_obj.create_challenge(options=options)
        # get details of token
        detail["transaction_ids"] = [c[2]]
        chal = {"transaction_id": c[2],
                "image": None,
                "client_mode": CLIENTMODE.INTERACTIVE,
                "serial": token_obj.token.serial,
                "type": token_obj.type,
                "message": message or _("Please enter your new email address!")}
        detail["multi_challenge"] = [chal]
        detail.update(chal)

    def enroll_via_validate_2nd_step(self, passw, options=None):
        """
        This method is the optional second step of ENROLL_VIA_MULTICHALLENGE.
        It is used in situations like the email token, sms token or push,
        when enrollment via challenge response needs two steps.

        The passw is entered during the first authentication step and it
        contains the email address.

        So we need to update the token with the email address and
        we need to create a new challenge for the final authentication.

        :param options:
        :return:
        """
        # Get policy for email validation
        validate_module = "privacyidea.lib.utils.emailvalidation"
        if "g" in options:
            g = options.get("g")
            validate_modules = Match.user(g, scope=SCOPE.ENROLL,
                                          action=ACTION.EMAILVALIDATION,
                                          user_object=options.get("user") if "user" in options else None).action_values(
                unique=True)
            # check passw to be a valid email address
            if len(validate_modules) == 1:
                # We have a policy for email validation
                validate_module = list(validate_modules)[0]
        validate_email = get_email_validators().get(validate_module)
        if validate_email(passw):
            # TODO: If anything special happens, we could leave it as a dynamic email
            self.del_tokeninfo("dynamic_email")
            self.add_tokeninfo(self.EMAIL_ADDRESS_KEY, passw)
            # Dynamically we remember that we need to do another challenge
            self.currently_in_challenge = True
        else:
            self.token.delete()
            raise ValidateError(_("The email address is not valid!"))



#    2017-12-27 Cornelius Kölbel <cornelius.koelbel@netknights.i>
#               Restructuring Code.
#
#    E-mail: yurlov.alexandr@gmail.com
#
#    2017-12-22 Alexander Yurlov <yurlov.alexandr@gmail.com>
#               Add new SMPP SMS provider
#
#    This program is free software: you can redistribute it and/or
#    modify it under the terms of the GNU Affero General Public
#    License, version 3, as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the
#               GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

__doc__ = """This is the SMSClass to send SMS via SMPP protocol to SMS center
It requires smpplib installation, this lib works with ascii only, but message support unicode 

"""

from privacyidea.lib.smsprovider.SMSProvider import (ISMSProvider, SMSError)
from privacyidea.lib import _
from privacyidea.lib.utils import parse_int
import logging
import traceback
import smpplib
import smpplib.gsm

log = logging.getLogger(__name__)


class SmppSMSProvider(ISMSProvider):

    def submit_message(self, phone, message):
        """
        send a message to a phone via a smpp protocol to smsc

        :param phone: the phone number
        :param message: the message to submit to the phone
        :return:
        """
        if not self.smsgateway:
            # this should not happen. We now always use sms gateway definitions.
            log.warning("Missing smsgateway definition!")
            raise SMSError(-1, "Missing smsgateway definition!")

        phone = self._mangle_phone(phone, self.smsgateway.option_dict)
        log.debug("submitting message {0!r} to {1!s}".format(message, phone))

        smsc_host = self.smsgateway.option_dict.get("SMSC_HOST")
        smsc_port = self.smsgateway.option_dict.get("SMSC_PORT")
        sys_id = self.smsgateway.option_dict.get("SYSTEM_ID")
        passwd = self.smsgateway.option_dict.get("PASSWORD")
        s_addr_ton = parse_int(self.smsgateway.option_dict.get("S_ADDR_TON"))
        s_addr_npi = parse_int(self.smsgateway.option_dict.get("S_ADDR_NPI"))
        s_addr = self.smsgateway.option_dict.get("S_ADDR")
        d_addr_ton = parse_int(self.smsgateway.option_dict.get("D_ADDR_TON"))
        d_addr_npi = parse_int(self.smsgateway.option_dict.get("D_ADDR_NPI"))

        if not smsc_host:
            log.warning("Can not submit message. SMSC_HOST is missing.")
            raise SMSError(-1, "No SMSC_HOST specified in the provider config.")

        if not smsc_port:
            log.warning("Can not submit message. SMSC_PORT is missing.")
            raise SMSError(-1, "No SMSC_PORT specified in the provider config.")

        # Initialize the SMPP Client
        client = None
        error_message = None 
        try:
            msg_parts, encoding_flag, msg_type_flag = smpplib.gsm.make_parts(message)
            client = smpplib.client.Client(smsc_host,
                                           smsc_port)
            client.connect()
            r = client.bind_transmitter(system_id=sys_id,
                                        password=passwd)
            log.debug("bind_transmitter returns {0!r}".format(r.get_status_desc()))
            for part in msg_parts:
                r = client.send_message(source_addr_ton=s_addr_ton,
                                        source_addr_npi=s_addr_npi,
                                        source_addr=s_addr,
                                        dest_addr_ton=d_addr_ton,
                                        dest_addr_npi=d_addr_npi,
                                        destination_addr=phone,
                                        short_message=part,
                                        data_coding=encoding_flag,
                                        esm_class=msg_type_flag)
                log.debug("send_message returns {0!r}".format(r.get_status_desc()))

        except Exception as err:
            error_message = "{0!r}".format(err)
            log.warning("Failed to send message: {0!r}".format(error_message))
            log.debug("{0!s}".format(traceback.format_exc()))

        finally:
            if client:
                client.disconnect()

        if error_message:
            raise SMSError(error_message, "SMS could not be "
                                          "sent: {0!r}".format(error_message))
        return True

    @classmethod
    def parameters(cls):
        """
        Return a dictionary, that describes the parameters and options for the
        SMS provider.
        Parameters are required keys to values.

        :return: dict
        """
        params = {"options_allowed": False,
                  "headers_allowed": False,
                    "parameters": {
                        "SMSC_HOST": {
                            "required": True,
                            "description": _("SMSC Host IP")},
                        "SMSC_PORT": {
                            "required": True,
                            "description": _("SMSC Port")},
                        "SYSTEM_ID": {
                            "description": _("SMSC Service ID")},
                        "PASSWORD": {
                            "description": _("Password for authentication on SMSC")},
                        "S_ADDR_TON": {
                            "description": _("SOURCE_ADDR_TON Special Flag")},
                        "S_ADDR_NPI": {
                            "description": _("S_ADDR_NPI Special Flag")},
                        "S_ADDR": {
                            "description": _("Source address (SMS sender)")},
                        "D_ADDR_TON": {"description": _("DESTINATION_ADDR_TON Special Flag")},
                        "D_ADDR_NPI": {"description": _("D_ADDR_NPI Special Flag")},
                        "REGEXP": {
                            "description": cls.regexp_description
                        }
                    }
                    }
        return params



#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
'''
  It generates the URL for smartphone apps like
                google authenticator
                oath token
  Implementation inspired by:
      https://github.com/neush/otpn900/blob/master/src/test_motp.c
'''

import logging
import time

from hashlib import md5

from privacyidea.lib.utils import to_unicode, to_bytes
from privacyidea.lib.crypto import zerome, safe_compare
from privacyidea.lib.log import log_with


log = logging.getLogger(__name__)


class mTimeOtp(object):
    '''
    implements the motp timebased check_otp
    - s. https://github.com/neush/otpn900/blob/master/src/test_motp.c
    '''

    def __init__(self, secObj=None, secPin=None, oldtime=0, digits=6,
                 key=None, pin=None):
        '''
        constructor for the mOtp class
        
        :param secObj:    secretObject, which covers the encrypted secret
        :param secPin:    secretObject, which covers the encrypted pin
        :param oldtime:   the previously detected otp counter/time
        :param digits:    length of otp chars to be tested
        :param key:       direct key provider (for selfTest)
        :param pin:       direct pin provider (for selfTest)
        
        :return:          nothing
        '''
        self.secretObject = secObj
        self.key = key

        self.secPin = secPin
        self.pin = pin

        self.oldtime = oldtime  ## last time access
        self.digits = digits

    @log_with(log)
    def checkOtp(self, anOtpVal, window=10, options=None):
        '''
        check a provided otp value

        :param anOtpVal: the to be tested otp value
        :type anOtpVal: str
        :param window: the +/- window around the test time
        :param options: generic container for additional values \
                        here only used for seltest: setting the initTime

        :return: -1 for fail else the identified counter/time 
        '''
        res = -1
        window = window * 2

        initTime = 0
        if options is not None and type(options) == dict:
            initTime = int(options.get('initTime', 0))

        if initTime == 0:
            otime = int(time.time() // 10)
        else:
            otime = initTime

        if self.secretObject is None:
            key = self.key
            pin = self.pin
        else:
            key = self.secretObject.getKey()
            pin = self.secPin.getKey()

        for i in range(otime - window, otime + window):
            otp = self.calcOtp(i, to_unicode(key), to_unicode(pin))
            if safe_compare(anOtpVal, otp):
                res = i
                log.debug("otpvalue {0!r} found at: {1!r}".format(anOtpVal, res))
                break

        if self.secretObject is not None:
            zerome(key)
            zerome(pin)
            del key
            del pin

        ## prevent access twice with last motp
        if res <= self.oldtime:
            log.warning("otpvalue {0!s} checked once before ({1!r}<={2!r})".format(anOtpVal, res, self.oldtime))
            res = -1
        if res == -1:
            msg = 'checking motp failed'
        else:
            msg = 'checking motp sucess'

        log.debug("end. {0!s} : returning result: {1!r}, ".format(msg, res))
        return res

    def calcOtp(self, counter, key=None, pin=None):
        '''
        calculate an otp value from counter/time, key and pin
        
        :param counter:    counter/time to be checked
        :type counter:     int
        :param key:        the secret key
        :type key:         str
        :param pin:        the secret pin
        :type pin:         str
        
        :return:           the otp value
        :rtype:            str
        '''
        ## ref impl from https://github.com/szimszon/motpy/blob/master/motpy
        if pin is None:
            pin = self.pin
        if key is None:
            key = self.key

        vhash = "{0:d}{1!s}{2!s}".format(counter, key, pin)
        motp = md5(to_bytes(vhash)).hexdigest()[:self.digits]  # nosec B324 # md5 used in mOTP algorithm
        return to_unicode(motp)



#  privacyIDEA
#  Jul 18, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
These are the applications that can be assigned to machines in privacyIDEA.

An application must be inherited from the MachineApplicationBase class.
"""
from .base import MachineApplication as MachineApplicationBase
from .base import get_auth_item
from .base import is_application_allow_bulk_call
from .base import get_application_types



#  2017-08-11 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add condition for detail->error->message
#  2017-07-19 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add possibility to compare tokeninfo field against fixed time
#             and also {now} with offset.
#  2016-05-04 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Initial writeup
#
# License:  AGPLv3
# (c) 2016. Cornelius Kölbel
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """This is the base class for an event handler module.
The event handler module is bound to an event together with

* a condition and
* an action
* optional options ;-)
"""

import datetime
import logging
import re

from dateutil.tz import tzlocal

from privacyidea.lib import _
from privacyidea.lib.auth import ROLE
from privacyidea.lib.config import get_token_types
from privacyidea.lib.container import (find_container_by_serial, find_container_for_token, get_all_containers,
                                       get_container_classes, get_container_realms)
from privacyidea.lib.containerclass import TokenContainerClass
from privacyidea.lib.counter import read as counter_read
from privacyidea.lib.realm import get_realms
from privacyidea.lib.resolver import get_resolver_list
from privacyidea.lib.token import get_token_owner, get_tokens
from privacyidea.lib.user import User
from privacyidea.lib.utils import (compare_condition, compare_generic_condition,
                                   parse_time_offset_from_now, is_true,
                                   check_ip_in_policy, AUTH_RESPONSE, compare_time)
from privacyidea.lib.tokenclass import DATE_FORMAT
from privacyidea.lib.challenge import get_challenges

log = logging.getLogger(__name__)


class CONDITION(object):
    """
    Possible conditions
    """
    TOKEN_HAS_OWNER = "token_has_owner"  # nosec B105 # condition name
    TOKEN_IS_ORPHANED = "token_is_orphaned"  # nosec B105 # condition name
    TOKEN_LOCKED = "token_locked"  # nosec B105 # condition name
    TOKEN_VALIDITY_PERIOD = "token_validity_period"  # nosec B105 # condition name
    USER_TOKEN_NUMBER = "user_token_number"  # nosec B105 # condition name
    USER_CONTAINER_NUMBER = "user_container_number"
    OTP_COUNTER = "otp_counter"
    TOKENTYPE = "tokentype"
    LAST_AUTH = "last_auth"
    COUNT_AUTH = "count_auth"
    COUNT_AUTH_SUCCESS = "count_auth_success"
    COUNT_AUTH_FAIL = "count_auth_fail"
    COUNTER = "counter"
    FAILCOUNTER = 'failcounter'
    TOKENINFO = "tokeninfo"
    DETAIL_ERROR_MESSAGE = "detail_error_message"
    DETAIL_MESSAGE = "detail_message"
    RESULT_VALUE = "result_value"
    RESULT_STATUS = "result_status"
    RESULT_AUTHENTICATION = "result_authentication"
    TOKENREALM = "tokenrealm"
    TOKENRESOLVER = "tokenresolver"
    TOKEN_IS_IN_CONTAINER = "token_is_in_container"
    REALM = "realm"
    RESOLVER = "resolver"
    CLIENT_IP = "client_ip"
    ROLLOUT_STATE = "rollout_state"
    CHALLENGE_SESSION = "challenge_session"
    CHALLENGE_EXPIRED = "challenge_expired"
    CONTAINER_STATE = "container_state"
    CONTAINER_EXACT_STATE = "container_exact_state"
    CONTAINER_HAS_OWNER = "container_has_owner"
    CONTAINER_TYPE = "container_type"
    CONTAINER_HAS_TOKEN = "container_has_token"
    SERIAL = "serial"
    CONTAINER_INFO = "container_info"
    CONTAINER_REALM = "container_realm"
    CONTAINER_RESOLVER = "container_resolver"
    CONTAINER_LAST_AUTH = "container_last_authentication"
    CONTAINER_LAST_SYNC = "container_last_synchronization"


class GROUP(object):
    """
    These are the event handler groups. The conditions
    will be grouped in the UI.
    """
    TOKEN = "token"  # nosec B105 # group name
    GENERAL = "general"
    USER = "user"
    COUNTER = "counter"
    CHALLENGE = "challenge"
    CONTAINER = "container"


class BaseEventHandler(object):
    """
    An Eventhandler needs to return a list of actions, which it can handle.

    It also returns a list of allowed action and conditions

    It returns an identifier, which can be used in the event-handling definitions
    """

    identifier = "BaseEventHandler"
    description = "This is the base class of an EventHandler with no " \
                  "functionality"
    run_details = None

    def __init__(self):
        pass

    @property
    def allowed_positions(self):
        """
        This returns the allowed positions of the event handler definition.
        This can be "post" or "pre" or both.

        :return: list of allowed positions
        """
        return ["post"]

    @property
    def actions(self):
        """
        This method returns a list of available actions, that are provided
        by this event handler.

        :return: dictionary of actions.
        """
        actions = ["sample_action_1", "sample_action_2"]
        return actions

    @property
    def conditions(self):
        """
        The UserNotification can filter for conditions like
        * type of logged-in user and
        * successful or failed value.success

        allowed types are str, multi, text, regexp

        :return: dict
        """
        realms = get_realms()
        resolvers = get_resolver_list()
        container_states = [{"name": state} for state in TokenContainerClass.get_state_types().keys()]
        cond = {
            CONDITION.CHALLENGE_SESSION:
                {
                    "type": "str",
                    "desc": _("The challenge session matches the string or regular "
                              "expression (like 'challenge_declined' or 'enrollment')"),
                    "group": GROUP.CHALLENGE
                },
            CONDITION.CHALLENGE_EXPIRED:
                {
                    "type": "str",
                    "desc": _("The challenge of a token during the authentication process"
                              " is expired."),
                    "value": ("True", "False"),
                    "group": GROUP.CHALLENGE
                },
            CONDITION.ROLLOUT_STATE:
                {
                    "type": "str",
                    "desc": _("The rollout_state of the token has a certain value like 'clientwait' or 'enrolled'."),
                    "group": GROUP.TOKEN
                },
            CONDITION.REALM:
                {
                    "type": "multi",
                    "desc": _("The realm of the user, for which this event should apply."),
                    "value": [{"name": r} for r in realms],
                    "group": GROUP.USER
                },
            CONDITION.RESOLVER:
                {
                    "type": "multi",
                    "desc": _("The resolver of the user, for which this event should apply."),
                    "value": [{"name": r} for r in resolvers],
                    "group": GROUP.USER
                },
            CONDITION.TOKENREALM:
                {
                    "type": "multi",
                    "desc": _("The realm of the token, for which this event should "
                              "apply."),
                    "value": [{"name": r} for r in realms],
                    "group": GROUP.TOKEN
                },
            CONDITION.TOKENRESOLVER:
                {
                    "type": "multi",
                    "desc": _("The resolver of the token, for which this event should "
                              "apply."),
                    "value": [{"name": r} for r in resolvers],
                    "group": GROUP.TOKEN
                },
            CONDITION.TOKENTYPE:
                {
                    "type": "multi",
                    "desc": _("The type of the token."),
                    "value": [{"name": r} for r in get_token_types()],
                    "group": GROUP.TOKEN
                },
            "logged_in_user":
                {
                    "type": "str",
                    "desc": _("The logged in user is of the following type."),
                    "value": (ROLE.ADMIN, ROLE.USER),
                    "group": GROUP.USER
                },
            CONDITION.RESULT_VALUE:
                {
                    "type": "str",
                    "desc": _("The result.value within the response is "
                              "True or False."),
                    "value": ("True", "False"),
                    "group": GROUP.GENERAL
                },
            CONDITION.RESULT_STATUS:
                {
                    "type": "str",
                    "desc": _("The result.status within the response is "
                              "True or False."),
                    "value": ("True", "False"),
                    "group": GROUP.GENERAL
                },
            CONDITION.RESULT_AUTHENTICATION:
                {
                    "type": "str",
                    "desc": _("The result.authentication within the response is the given value."),
                    "value": (
                        AUTH_RESPONSE.ACCEPT, AUTH_RESPONSE.REJECT, AUTH_RESPONSE.CHALLENGE, AUTH_RESPONSE.DECLINED),
                    "group": GROUP.GENERAL
                },
            CONDITION.TOKEN_LOCKED:
                {
                    "type": "str",
                    "desc": _("Check if the max failcounter of the token is "
                              "reached."),
                    "value": ("True", "False"),
                    "group": GROUP.TOKEN
                },
            CONDITION.TOKEN_HAS_OWNER:
                {
                    "type": "str",
                    "desc": _("The token has a user assigned."),
                    "value": ("True", "False"),
                    "group": GROUP.TOKEN
                },
            CONDITION.TOKEN_IS_ORPHANED:
                {
                    "type": "str",
                    "desc": _("The token has a user assigned, but the user does "
                              "not exist in the userstore anymore."),
                    "value": ("True", "False"),
                    "group": GROUP.TOKEN
                },
            CONDITION.TOKEN_VALIDITY_PERIOD:
                {
                    "type": "str",
                    "desc": _("Check if the token is within its validity period."),
                    "value": ("True", "False"),
                    "group": GROUP.TOKEN
                },
            CONDITION.SERIAL:
                {
                    "type": "regexp",
                    "desc": _("Action is triggered, if the serial matches this "
                              "regular expression."),
                    "group": GROUP.TOKEN
                },
            CONDITION.TOKEN_IS_IN_CONTAINER:
                {
                    "type": "str",
                    "desc": _("The token is in a container."),
                    "value": ("True", "False"),
                    "group": GROUP.TOKEN
                },
            CONDITION.USER_TOKEN_NUMBER:
                {
                    "type": "str",
                    "desc": _("Action is triggered, if the user has this number "
                              "of tokens assigned. > and < can be used, like <8."),
                    "group": GROUP.USER
                },
            CONDITION.USER_CONTAINER_NUMBER:
                {
                    "type": "str",
                    "desc": _("Action is triggered, if the user has this number "
                              "of containers assigned."),
                    "group": GROUP.USER
                },
            CONDITION.OTP_COUNTER:
                {
                    "type": "str",
                    "desc": _("Action is triggered, if the counter of the token "
                              "equals this setting. Can also be "
                              "'>100' or '<99' for no exact match."),
                    "group": GROUP.COUNTER
                },
            CONDITION.LAST_AUTH:
                {
                    "type": "str",
                    "desc": _("Action is triggered, if the last authentication of "
                              "the token is older than 7h, 10d or 1y."),
                    "group": GROUP.TOKEN
                },
            CONDITION.COUNT_AUTH:
                {
                    "type": "str",
                    "desc": _("This can be '>100', '<99', or '=100', to trigger "
                              "the action, if the tokeninfo field 'count_auth' is "
                              "bigger than 100, less than 99 or exactly 100."),
                    "group": GROUP.COUNTER
                },
            CONDITION.COUNT_AUTH_SUCCESS:
                {
                    "type": "str",
                    "desc": _("This can be '>100', '<99', or '=100', to trigger "
                              "the action, if the tokeninfo field "
                              "'count_auth_success' is "
                              "bigger than 100, less than 99 or exactly 100."),
                    "group": GROUP.COUNTER
                },
            CONDITION.COUNT_AUTH_FAIL:
                {
                    "type": "str",
                    "desc": _("This can be '>100', '<99', or '=100', to trigger "
                              "the action, if the difference between the tokeninfo "
                              "field 'count_auth' and 'count_auth_success is "
                              "bigger than 100, less than 99 or exactly 100."),
                    "group": GROUP.COUNTER
                },
            CONDITION.FAILCOUNTER:
                {
                    "type": "str",
                    "desc": _("This can be '>9', '<9', or '=10', to trigger "
                              "the action, if the failcounter of a token matches this value. "
                              "Note that the failcounter stops increasing, if the max_failcount is "
                              "reached."),
                    "group": GROUP.COUNTER
                },
            CONDITION.TOKENINFO:
                {
                    "type": "str",
                    "desc": _("This condition can check any arbitrary tokeninfo "
                              "field. You need to enter something like "
                              "'<fieldname> == <fieldvalue>', '<fieldname> > "
                              "<fieldvalue>' or '<fieldname> < <fieldvalue>'."),
                    "group": GROUP.TOKEN
                },
            CONDITION.COUNTER:
                {
                    "type": "str",
                    "desc": _("This condition can check the value of an arbitrary event counter and "
                              "compare it like 'myCounter == 1000', 'myCounter > 1000' or "
                              "'myCounter < 1000'."),
                    "group": GROUP.COUNTER
                },
            CONDITION.DETAIL_ERROR_MESSAGE:
                {
                    "type": "str",
                    "desc": _("Here you can enter a regular expression. The "
                              "condition only applies if the regular expression "
                              "matches the detail->error->message in the response."),
                    "group": GROUP.GENERAL
                },
            CONDITION.DETAIL_MESSAGE:
                {
                    "type": "str",
                    "desc": _("Here you can enter a regular expression. The "
                              "condition only applies if the regular expression "
                              "matches the detail->message in the response."),
                    "group": GROUP.GENERAL
                },
            CONDITION.CLIENT_IP:
                {
                    "type": "str",
                    "desc": _("Trigger the action, if the client IP matches."),
                    "group": GROUP.GENERAL
                },
            CONDITION.CONTAINER_STATE:
                {
                    "type": "multi",
                    "desc": _("The container is in the specified states, but can additionally be in other states."),
                    "value": container_states,
                    "group": GROUP.CONTAINER
                },
            CONDITION.CONTAINER_EXACT_STATE:
                {
                    "type": "multi",
                    "desc": _("The container is only in the specified states."),
                    "value": container_states,
                    "group": GROUP.CONTAINER
                },
            CONDITION.CONTAINER_HAS_OWNER:
                {
                    "type": "str",
                    "desc": _("The container has a user assigned."),
                    "value": ("True", "False"),
                    "group": GROUP.CONTAINER
                },
            CONDITION.CONTAINER_HAS_TOKEN:
                {
                    "type": "str",
                    "desc": _("The container has at least one token assigned."),
                    "value": ("True", "False"),
                    "group": GROUP.CONTAINER
                },
            CONDITION.CONTAINER_TYPE:
                {
                    "type": "str",
                    "desc": _("The container is of a certain type."),
                    "value": list(get_container_classes().keys()),
                    "group": GROUP.CONTAINER
                },
            CONDITION.CONTAINER_REALM:
                {
                    "type": "multi",
                    "desc": _("The container is in this realm or in no realm at all. If multiple realms are selected, "
                              "the condition is fulfilled if the container is in at least one realm of the list. The "
                              "condition is not checked if the container has no realm, hence the action would be "
                              "triggered."),
                    "value": [{"name": r} for r in realms],
                    "group": GROUP.CONTAINER
                },
            CONDITION.CONTAINER_RESOLVER:
                {
                    "type": "multi",
                    "desc": _("An owner of the container is in this resolver. If multiple resolvers are selected, the "
                              "condition is fulfilled if at least one owner is in one resolver. The condition is not "
                              "checked if the container has no owner, hence the action would be triggered."),
                    "value": [{"name": r} for r in resolvers],
                    "group": GROUP.CONTAINER
                },
            CONDITION.CONTAINER_INFO:
                {
                    "type": "str",
                    "desc": _("This condition can check any arbitrary container info field. You need to enter "
                              "something like 'fieldname == fieldvalue', 'fieldname > fieldvalue' or 'fieldname < "
                              "fieldvalue'."),
                    "group": GROUP.CONTAINER
                },
            CONDITION.CONTAINER_LAST_AUTH:
                {
                    "type": "str",
                    "desc": _(
                        "Action is triggered, if the last authentication of the container is older than the specified "
                        "time. The time value has to be an integer followed by a time unit. Supported units are "
                        "'y' (years), 'd' (days), 'h' (hours), 'm' (minutes), 's' (seconds). Only one unit is allowed. "
                        "Examples: '8h', '7d', '1y'"),

                    "group": GROUP.CONTAINER
                },
            CONDITION.CONTAINER_LAST_SYNC:
                {
                    "type": "str",
                    "desc": _(
                        "Action is triggered, if the last synchronization of the container is older than the specified "
                        "time. The time value has to be an integer followed by a time unit. Supported units are "
                        "'y' (years), 'd' (days), 'h' (hours), 'm' (minutes), 's' (seconds). Only one unit is allowed. "
                        "Examples: '8h', '7d', '1y'"),
                    "group": GROUP.CONTAINER
                },
        }
        return cond

    @property
    def events(self):
        """
        This method returns a list allowed events, that this event handler
        can be bound to and which it can handle with the corresponding actions.

        An eventhandler may return an asterisk ["*"] indicating, that it can
        be used in all events.

        :return: list of events
        """
        events = ["*"]
        return events

    @staticmethod
    def _get_tokenowner(request):
        user = User()
        if hasattr(request, "User"):
            user = request.User
            if user is None:
                user = User()
            serial = request.all_data.get("serial")
            if user.is_empty() and serial:
                # maybe the user is empty, but a serial was passed.
                # Then we determine the user by the serial
                try:
                    user = get_token_owner(serial) or User()
                except Exception as exx:
                    user = User()
                    # This can happen for orphaned tokens.
                    log.info("Could not determine tokenowner for {0!s}. Maybe the "
                             "user does not exist anymore.".format(serial))
                    log.debug(exx)
            # If the user does not exist, we set an empty user
            if not user.exist():
                user = User()

        return user

    @staticmethod
    def _get_container_owners(request):
        users = []
        user = User(login='', realm='')
        if hasattr(request, "User"):
            user = request.User
            if user:
                users.append(user)
        serial = request.all_data.get("container_serial")
        if not user and serial:
            # maybe the user is empty, but a serial was passed.
            # Then we determine the user by the serial
            container = find_container_by_serial(serial)
            users = container.get_users()

        return users

    @classmethod
    def _get_users_from_request(cls, request):
        """
        Extracts the user information from the request or searches for the container owners or token owner.
        If no user is found, an empty user is returned.

        :param request: The request object
        :return: List of user objects
        """
        users = cls._get_container_owners(request)
        if len(users) == 0:
            user = cls._get_tokenowner(request)
            if not user.is_empty():
                users.append(user)
        return users

    @classmethod
    def _get_token_serials(cls, request, content, g):
        """
        Extracts the token serials from the request, content or audit object.

        :param request: The request object
        :param content: The content of the response
        :param g: The g object
        :return: Comma separated string of token serials
        """
        # Get single token serial
        serial = request.all_data.get("serial") or \
                 content.get("detail", {}).get("serial") or \
                 g.audit_object.audit_data.get("serial")

        return serial

    @classmethod
    def _get_container_serial(cls, request, content):
        """
        Get the container serial from the request, content or audit object.

        :param request: The request object
        :param content: The content of the response
        :return: The container serial or None if no serial could be found
        """

        # get serial from request
        container_serial = request.all_data.get("container_serial")
        if not cls._serial_is_from_container(container_serial):
            container_serial = None

        if not container_serial:
            # get serial from response
            value = content.get("result", {}).get('value', {})
            if isinstance(value, dict):
                container_serial = value.get("container_serial")
                if not cls._serial_is_from_container(container_serial):
                    container_serial = None

        return container_serial

    @classmethod
    def _get_container_serial_from_token(cls, request, content, g):
        """
        Tries to get a token serial. If a token is found, it tries to get the container serial from this token.

        :param request: The request object
        :param content: The content of the response
        :param g: The g object
        :return: The container serial or None if no serial could be found
        """
        container_serial = None
        serials = cls._get_token_serials(request, content, g)

        if serials:
            serial_list = serials.replace(' ', '').split(',')
            if len(serial_list):
                # Only if one token serial is provided a corresponding container can be found
                serial = serial_list[0]
                container_serial = cls._get_container_serial_from_token_serial(serial)

        return container_serial

    @classmethod
    def _serial_is_from_container(cls, serial):
        """
        Validates whether the given serial is from a container.

        :param serial: The serial to validate
        :return: True if a container with the given serial exists, False otherwise
        """
        container_exists = False

        if serial:
            # Check if a container exists with this serial
            containers = get_all_containers(serial=serial, pagesize=0, page=0)['containers']
            if len(containers) > 0:
                container_exists = True

        return container_exists

    @classmethod
    def _get_container_serial_from_token_serial(cls, token_serial):
        """
        Tries to get the container serial from a token.

        :param token_serial: Serial of a token
        :return: container serial or None if the token is not part of a container or the token does not exist
        """
        container_serial = None
        if token_serial:
            tokens = get_tokens(serial=token_serial)
            if len(tokens) > 0:
                token = tokens[0]
                container = find_container_for_token(token.get_serial())
                if container:
                    container_serial = container.serial

        return container_serial

    @staticmethod
    def _get_response_content(response):
        content = {}
        if response:
            if response.is_json:
                content = response.json
        return content

    def check_condition(self, options):
        """
        Check if all conditions are met and if the action should be executed.
        If the conditions are met, we return "True"

        :return: True
        """
        g = options.get("g")
        request = options.get("request")
        response = options.get("response")
        e_handler_def = options.get("handler_def")
        if not e_handler_def:
            # options is the handler definition
            return True
        # conditions can be corresponding to the property conditions
        conditions = e_handler_def.get("conditions")
        content = self._get_response_content(response)
        user = self._get_tokenowner(request)

        serial = request.all_data.get("serial") or content.get("detail", {}).get("serial")
        transaction_id = request.all_data.get("transaction_id")
        tokenrealms = []
        tokenresolvers = []
        tokentype = None
        token_obj = None
        if serial:
            # We have determined the serial number from the request.
            token_obj_list = get_tokens(serial=serial)
        elif user:
            # We have to determine the token via the user object. But only if
            #  the user has only one token
            token_obj_list = get_tokens(user=user)
        else:
            token_obj_list = []

        if len(token_obj_list) == 1:
            # There is a token involved, so we determine its resolvers and realms
            token_obj = token_obj_list[0]
            tokenrealms = token_obj.get_realms()
            tokentype = token_obj.get_tokentype()

            all_realms = get_realms()
            for tokenrealm in tokenrealms:
                resolvers = all_realms.get(tokenrealm, {}).get("resolver", {})
                tokenresolvers.extend([r.get("name") for r in resolvers])
            tokenresolvers = list(set(tokenresolvers))

        # Get container
        container_serial = self._get_container_serial(request, content)
        container = None
        if container_serial:
            container = find_container_by_serial(container_serial)
        elif serial and token_obj:
            container = find_container_for_token(serial)

        if CONDITION.CLIENT_IP in conditions:
            if g and g.client_ip:
                ip_policy = [ip.strip() for ip in conditions.get(CONDITION.CLIENT_IP).split(",")]
                found, excluded = check_ip_in_policy(g.client_ip, ip_policy)
                if not found or excluded:
                    return False

        if CONDITION.REALM in conditions:
            if user.realm not in conditions.get(CONDITION.REALM).split(","):
                return False

        if CONDITION.RESOLVER in conditions:
            if user.resolver not in conditions.get(CONDITION.RESOLVER).split(","):
                return False

        if "logged_in_user" in conditions:
            # Determine the role of the user
            try:
                logged_in_user = g.logged_in_user
                user_role = logged_in_user.get("role")
            except Exception:
                # A non-logged-in-user is a User, not an admin
                user_role = ROLE.USER
            if user_role != conditions.get("logged_in_user"):
                return False

        if CONDITION.RESULT_VALUE in conditions:
            condition_value = conditions.get(CONDITION.RESULT_VALUE)
            result_value = content.get("result", {}).get("value")
            if is_true(condition_value) != is_true(result_value):
                return False

        if CONDITION.RESULT_STATUS in conditions:
            condition_value = conditions.get(CONDITION.RESULT_STATUS)
            result_status = content.get("result", {}).get("status")
            if is_true(condition_value) != is_true(result_status):
                return False

        if CONDITION.RESULT_AUTHENTICATION in conditions:
            condition_value = conditions.get(CONDITION.RESULT_AUTHENTICATION)
            result_auth = content.get("result", {}).get("authentication")
            if condition_value != result_auth:
                return False

        # checking of max-failcounter state of the token
        if  CONDITION.TOKEN_LOCKED in conditions:
            if token_obj:
                locked = token_obj.get_failcount() >= \
                         token_obj.get_max_failcount()
                if (conditions.get(CONDITION.TOKEN_LOCKED) in ["True", True]) != \
                        locked:
                    return False
            else:
                # check all tokens of the user, if any token is maxfail
                token_objects = get_tokens(user=user, maxfail=True)
                if not ','.join([tok.get_serial() for tok in token_objects]):
                    return False

        if CONDITION.TOKENREALM in conditions and tokenrealms:
            res = False
            for trealm in tokenrealms:
                if trealm in conditions.get(CONDITION.TOKENREALM).split(","):
                    res = True
                    break
            if not res:
                return False

        if CONDITION.TOKENRESOLVER in conditions and tokenresolvers:
            res = False
            for tres in tokenresolvers:
                if tres in conditions.get(CONDITION.TOKENRESOLVER).split(","):
                    res = True
                    break
            if not res:
                return False

        if "serial" in conditions and serial:
            serial_match = conditions.get("serial")
            if not bool(re.match(serial_match, serial)):
                return False

        if CONDITION.USER_TOKEN_NUMBER in conditions and user:
            num_tokens = get_tokens(user=user, count=True)
            cond = conditions.get(CONDITION.USER_TOKEN_NUMBER)
            if not compare_condition(cond, num_tokens):
                return False

        if CONDITION.USER_CONTAINER_NUMBER in conditions and user:
            container_list = get_all_containers(user=user, page=1, pagesize=1)
            num_containers = container_list['count']
            if num_containers != int(conditions.get(CONDITION.USER_CONTAINER_NUMBER)):
                return False

        if CONDITION.DETAIL_ERROR_MESSAGE in conditions:
            message = content.get("detail", {}).get("error", {}).get("message", "")
            search_exp = conditions.get(CONDITION.DETAIL_ERROR_MESSAGE)
            m = re.search(search_exp, message)
            if not bool(m):
                return False

        if CONDITION.DETAIL_MESSAGE in conditions:
            message = content.get("detail", {}).get("message", "")
            search_exp = conditions.get(CONDITION.DETAIL_MESSAGE)
            m = re.search(search_exp, message)
            if not bool(m):
                return False

        if CONDITION.COUNTER in conditions:
            # Can be counter==1000
            if not compare_generic_condition(conditions.get(CONDITION.COUNTER),
                                             lambda x: counter_read(x) or 0,
                                             "Misconfiguration in your counter "
                                             "condition: {0!s}"
                                             ):
                return False

        # Token specific conditions
        if token_obj:
            if CONDITION.TOKENTYPE in conditions:
                if tokentype not in conditions.get(CONDITION.TOKENTYPE).split(
                        ","):
                    return False

            if CONDITION.TOKEN_HAS_OWNER in conditions:
                uid = token_obj.get_user_id()
                check = conditions.get(CONDITION.TOKEN_HAS_OWNER)
                if uid and check in ["True", True]:
                    res = True
                elif not uid and check in ["False", False]:
                    res = True
                else:
                    log.debug("Condition token_has_owner for token {0!r} "
                              "not fulfilled.".format(token_obj))
                    return False

            if CONDITION.TOKEN_IS_ORPHANED in conditions:
                orphaned = token_obj.is_orphaned()
                check = conditions.get(CONDITION.TOKEN_IS_ORPHANED)
                if orphaned and check in ["True", True]:
                    res = True
                elif not orphaned and check in ["False", False]:
                    res = True
                else:
                    log.debug("Condition token_is_orphaned for token {0!r} not "
                              "fulfilled.".format(token_obj))
                    return False

            if CONDITION.TOKEN_VALIDITY_PERIOD in conditions:
                valid = token_obj.check_validity_period()
                if (conditions.get(CONDITION.TOKEN_VALIDITY_PERIOD) in ["True", True]) != valid:
                    return False

            if CONDITION.OTP_COUNTER in conditions:
                cond = conditions.get(CONDITION.OTP_COUNTER)
                if not compare_condition(cond, token_obj.token.count):
                    return False

            if CONDITION.LAST_AUTH in conditions:
                if token_obj.check_last_auth_newer(conditions.get(CONDITION.LAST_AUTH)):
                    return False

            if CONDITION.COUNT_AUTH in conditions:
                count = token_obj.get_count_auth()
                cond = conditions.get(CONDITION.COUNT_AUTH)
                if not compare_condition(cond, count):
                    return False

            if CONDITION.COUNT_AUTH_SUCCESS in conditions:
                count = token_obj.get_count_auth_success()
                cond = conditions.get(CONDITION.COUNT_AUTH_SUCCESS)
                if not compare_condition(cond, count):
                    return False

            if CONDITION.COUNT_AUTH_FAIL in conditions:
                count = token_obj.get_count_auth()
                c_success = token_obj.get_count_auth_success()
                c_fail = count - c_success
                cond = conditions.get(CONDITION.COUNT_AUTH_FAIL)
                if not compare_condition(cond, c_fail):
                    return False

            if CONDITION.FAILCOUNTER in conditions:
                failcount = token_obj.get_failcount()
                cond = conditions.get(CONDITION.FAILCOUNTER)
                res = compare_condition(cond, failcount)
                if not res:
                    return False

            if CONDITION.TOKENINFO in conditions:
                cond = conditions.get(CONDITION.TOKENINFO)
                # replace {now} in condition
                cond, td = parse_time_offset_from_now(cond)
                s_now = (datetime.datetime.now(tzlocal()) + td).strftime(
                    DATE_FORMAT)
                cond = cond.format(now=s_now)
                if not compare_generic_condition(cond,
                                                 token_obj.get_tokeninfo,
                                                 "Misconfiguration in your tokeninfo "
                                                 "condition: {0!s}"):
                    return False

            if CONDITION.ROLLOUT_STATE in conditions:
                cond = conditions.get(CONDITION.ROLLOUT_STATE)
                if not cond == token_obj.token.rollout_state:
                    return False

            if CONDITION.TOKEN_IS_IN_CONTAINER in conditions:
                cond = conditions.get(CONDITION.TOKEN_IS_IN_CONTAINER)
                container = find_container_for_token(serial)
                token_is_in_container = container is not None
                if token_is_in_container and cond in ["True", True]:
                    res = True
                elif not token_is_in_container and cond in ["False", False]:
                    res = True
                else:
                    log.debug(f"Condition {CONDITION.TOKEN_IS_IN_CONTAINER} for token {token_obj} "
                              "not fulfilled.")
                    return False

        # Evaluates the challenge either for a token or a container object
        if token_obj or container:
            serial = token_obj.get_serial() if token_obj else None
            if not serial:
                serial = container.serial
            if CONDITION.CHALLENGE_SESSION or CONDITION.CHALLENGE_EXPIRED in conditions:
                chals = get_challenges(serial=serial, transaction_id=transaction_id)
                if len(chals) == 1:
                    chal = chals[0]
                    if CONDITION.CHALLENGE_SESSION in conditions:
                        cond_match = conditions.get(CONDITION.CHALLENGE_SESSION)
                        if not bool(re.match(cond_match, chal.session)):
                            return False
                    if CONDITION.CHALLENGE_EXPIRED in conditions:
                        condition_value = conditions.get(CONDITION.CHALLENGE_EXPIRED)
                        if is_true(condition_value) == chal.is_valid():
                            return False
                elif len(chals) > 1:
                    # If there is more than one challenge, the conditions seams awkward
                    if token_obj:
                        log.warning(f"There is more than one challenge for token {token_obj.token.serial} "
                                    f"and transaction_id {transaction_id}.")
                    else:
                        log.warning(f"There is more than one challenge for container {container.serial} "
                                    f"and transaction_id {transaction_id}.")
                    return False

        # Container specific conditions
        if container:
            if CONDITION.CONTAINER_STATE in conditions:
                cond = conditions.get(CONDITION.CONTAINER_STATE).split(',')
                container_states = container.get_states()
                for cond_state in cond:
                    if cond_state not in container_states:
                        log.debug(f"Condition container_state {cond_state} for container {container.serial} "
                                  "not fulfilled.")
                        return False

            if CONDITION.CONTAINER_EXACT_STATE in conditions:
                cond = conditions.get(CONDITION.CONTAINER_EXACT_STATE).split(',')
                container_states = container.get_states()
                if len(cond) != len(container_states):
                    log.debug(f"Condition container_single_state {cond} for container {container.serial} "
                              "not fulfilled.")
                    return False

                for cond_state in cond:
                    if cond_state not in container_states:
                        log.debug(f"Condition container_state {cond_state} for container {container.serial} "
                                  "not fulfilled.")
                        return False

            if CONDITION.CONTAINER_HAS_OWNER in conditions:
                has_container_owner = len(container.get_users()) > 0
                cond = conditions.get(CONDITION.CONTAINER_HAS_OWNER)
                if has_container_owner and cond in ["True", True]:
                    res = True
                elif not has_container_owner and cond in ["False", False]:
                    res = True
                else:
                    log.debug(f"Condition container_has_owner for container {container.serial} "
                              "not fulfilled.")
                    return False

            if CONDITION.CONTAINER_TYPE in conditions:
                cond = conditions.get(CONDITION.CONTAINER_TYPE)
                if container.type != cond:
                    log.debug(f"Condition container_type {cond} for container {container.serial} "
                              "not fulfilled.")
                    return False

            if CONDITION.CONTAINER_HAS_TOKEN in conditions:
                tokens = [token.get_serial() for token in container.get_tokens()]
                cond = conditions.get(CONDITION.CONTAINER_HAS_TOKEN)
                if len(tokens) > 0 and cond in ["True", True]:
                    res = True
                elif len(tokens) == 0 and cond in ["False", False]:
                    res = True
                else:
                    log.debug(f"Condition container_has_token for container {container.serial} "
                              "not fulfilled.")
                    return False

            if CONDITION.CONTAINER_REALM in conditions:
                condition_realms = conditions.get(CONDITION.CONTAINER_REALM).split(",")
                container_realms = get_container_realms(container.serial)
                # If the container is in no realm, the condition is not checked, it is fulfilled
                if len(container_realms) > 0:
                    # check if at least one realm of the condition is also a realm of the container
                    matching_realms = list(set(condition_realms).intersection(container_realms))
                    if len(matching_realms) == 0:
                        log.debug(f"Condition container_realm {condition_realms} for container {container.serial} "
                                  "not fulfilled.")
                        return False

            if CONDITION.CONTAINER_RESOLVER in conditions:
                condition_resolvers = conditions.get(CONDITION.CONTAINER_RESOLVER).split(",")
                container_resolvers = [user.resolver for user in container.get_users()]
                # If the container has no owner, the condition is not checked, it is fulfilled
                if len(container_resolvers) > 0:
                    # check if at least one resolver of the condition is also a resolver of a container owner
                    matching_resolvers = list(set(condition_resolvers).intersection(container_resolvers))
                    if len(matching_resolvers) == 0:
                        log.debug(
                            f"Condition container_resolver {condition_resolvers} for container {container.serial} "
                            "not fulfilled.")
                        return False

            if CONDITION.CONTAINER_INFO in conditions:
                cond = conditions.get(CONDITION.CONTAINER_INFO)

                if not compare_generic_condition(cond,
                                                 container.get_container_info_dict().get,
                                                 "Misconfiguration in your container info condition: {0!s}"):
                    return False

            if CONDITION.CONTAINER_LAST_AUTH in conditions:
                cond = conditions.get(CONDITION.CONTAINER_LAST_AUTH)
                last_auth = container.last_authentication
                res = False

                if last_auth:
                    last_auth = last_auth.replace(tzinfo=datetime.timezone.utc)
                    res = compare_time(cond, last_auth)

                if not res:
                    return False

            if CONDITION.CONTAINER_LAST_SYNC in conditions:
                cond = conditions.get(CONDITION.CONTAINER_LAST_SYNC)
                last_sync = container.last_synchronization
                res = False

                if last_sync:
                    last_sync = last_sync.replace(tzinfo=datetime.timezone.utc)
                    res = compare_time(cond, last_sync)

                if not res:
                    return False

        return True

    def do(self, action, options=None):
        """
        This method executes the defined action in the given event.

        :param action:
        :param options: Contains the flask parameters g and request and the
            handler_def configuration
        :type options: dict
        :return:
        """
        log.info("In fact we are doing nothing, be we presume we are doing"
                 "{0!s}".format(action))
        return True



# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Jelina Unger <jelina.unger@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later

class SmartphoneOptions:
    """
    Options for the smartphone container.
    """
    KEY_ALGORITHM = "key_algorithm"
    HASH_ALGORITHM = "hash_algorithm"
    ENCRYPT_KEY_ALGORITHM = "encrypt_key_algorithm"
    ENCRYPT_ALGORITHM = "encrypt_algorithm"
    ENCRYPT_MODE = "encrypt_mode"



"""
This test file tests the lib.tokens.spasstoken
This depends on lib.tokenclass
"""

from .base import MyTestCase
from privacyidea.lib.tokens.spasstoken import SpassTokenClass
from privacyidea.models import Token


class SpassTokenTestCase(MyTestCase):

    otppin = "topsecret"
    serial1 = "ser1"

    # add_user, get_user, reset, set_user_identifiers
    
    def test_01_create_token(self):
        db_token = Token(self.serial1, tokentype="spass")
        db_token.save()
        token = SpassTokenClass(db_token)
        token.update({})
        self.assertTrue(token.token.serial == self.serial1, token)
        self.assertTrue(token.token.tokentype == "spass", token.token.tokentype)
        self.assertTrue(token.type == "spass", token)
        class_prefix = token.get_class_prefix()
        self.assertTrue(class_prefix == "PISP", class_prefix)
        self.assertTrue(token.get_class_type() == "spass", token)

    def test_02_check_password(self):
        db_token = Token.query.filter(Token.serial == self.serial1).first()
        token = SpassTokenClass(db_token)

        r = token.check_otp("")
        self.assertTrue(r == 0, r)

        r = token.check_otp("wrong pw")
        self.assertTrue(r == 0, r)

        # check pin+otp:
        token.set_pin(self.otppin)
        r = token.authenticate(self.otppin)
        self.assertTrue(r, r)

    def test_03_class_methods(self):
        db_token = Token.query.filter(Token.serial == self.serial1).first()
        token = SpassTokenClass(db_token)

        info = token.get_class_info()
        self.assertTrue(info.get("title") == "Simple Pass Token", info)

        info = token.get_class_info("title")
        self.assertTrue(info == "Simple Pass Token", info)



# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Nils Behlen <nils.behlen@netknights.it>
# SPDX-FileCopyrightText: 2024 Jelina Unger <jelina.unger@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
import logging

from privacyidea.lib import _
from privacyidea.lib.containerclass import TokenContainerClass

log = logging.getLogger(__name__)


class YubikeyContainer(TokenContainerClass):
    options = {}

    def __init__(self, db_container):
        super().__init__(db_container)

    @classmethod
    def get_class_type(cls) -> str:
        """
        Returns the type of the container class.
        """
        return "yubikey"

    @classmethod
    def get_supported_token_types(cls) -> list:
        """
        Returns the token types that are supported by the container class.
        """
        supported_token_types = ["hotp", "certificate", "webauthn", "yubico", "yubikey", "passkey"]
        supported_token_types.sort()
        return supported_token_types

    @classmethod
    def get_class_prefix(cls) -> str:
        """
        Returns the container class specific prefix for the serial.
        """
        return "YUBI"

    @classmethod
    def get_class_description(cls) -> str:
        """
        Returns a description of the container class.
        """
        return _("Yubikey hardware device that can hold HOTP, certificate and webauthn token")



# http://www.privacyidea.org
# (c) Cornelius Kölbel, privacyidea.org
#
# 2015-02-26 Cornelius Kölbel, <cornelius@privacyidea.org>
#            Initial writeup
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This endpoint is used to create, modify, list and delete Machine Resolvers.
Machine Resolvers fetch machine information from remote machine stores like a
hosts file or an Active Directory.

The code of this module is tested in tests/test_api_machineresolver.py
"""
from flask import (Blueprint,
                   request)
from .lib.utils import (getParam,
                        optional,
                        required,
                        send_result)
from ..lib.log import log_with
from ..lib.machineresolver import (get_resolver_list, save_resolver, delete_resolver,
                           pretestresolver)
from flask import g
import logging
from ..api.lib.prepolicy import prepolicy, check_base_action
from ..lib.policy import ACTION


log = logging.getLogger(__name__)


machineresolver_blueprint = Blueprint('machineresolver_blueprint', __name__)


@machineresolver_blueprint.route('/', methods=['GET'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.MACHINERESOLVERREAD)
def get_resolvers():
    """
    returns a json list of all machine resolver.

    :param type: Only return resolvers of type (like "hosts"...)
    """
    typ = getParam(request.all_data, "type", optional)
    res = get_resolver_list(filter_resolver_type=typ)
    g.audit_object.log({"success": True})
    return send_result(res)


@machineresolver_blueprint.route('/<resolver>', methods=['POST'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.MACHINERESOLVERWRITE)
def set_resolver(resolver=None):
    """
    This creates a new machine resolver or updates an existing one.
    A resolver is uniquely identified by its name.

    If you update a resolver, you do not need to provide all parameters.
    Parameters you do not provide are left untouched.
    When updating a resolver you must not change the type!
    You do not need to specify the type, but if you specify a wrong type,
    it will produce an error.

    :param resolver: the name of the resolver.
    :type resolver: basestring
    :param type: the type of the resolver. Valid types are... "hosts"
    :type type: string
    :return: a json result with the value being the database id (>0)

    Additional parameters depend on the resolver type.

    hosts:
     * filename
    """
    param = request.all_data
    if resolver:
        # The resolver parameter was passed as a part of the URL
        param.update({"name": resolver})
    res = save_resolver(param)
    return send_result(res)


@machineresolver_blueprint.route('/<resolver>', methods=['DELETE'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.MACHINERESOLVERDELETE)
def delete_resolver_api(resolver=None):
    """
    this function deletes an existing machine resolver

    :param resolver: the name of the resolver to delete.
    :return: json with success or fail
    """
    res = delete_resolver(resolver)
    g.audit_object.log({"success": res,
                        "info": resolver})

    return send_result(res)


@machineresolver_blueprint.route('/<resolver>', methods=['GET'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.MACHINERESOLVERREAD)
def get_resolver(resolver=None):
    """
    This function retrieves the definition of a single machine resolver.

    :param resolver: the name of the resolver
    :return: a json result with the configuration of a specified resolver
    """
    res = get_resolver_list(filter_resolver_name=resolver)

    g.audit_object.log({"success": True,
                        "info": resolver})

    return send_result(res)


@machineresolver_blueprint.route('/test', methods=["POST"])
@log_with(log)
def test_resolver():
    """
    This function tests, if the given parameter will create a working
    machine resolver. The Machine Resolver Class itself verifies the
    functionality. This can also be network connectivity to a Machine Store.

    :return: a json result with bool
    """
    param = request.all_data
    rtype = getParam(param, "type", required)
    success, desc = pretestresolver(rtype, param)
    return send_result(success, details={"description": desc})




"""v3.8: Add indices to datetime columns

Revision ID: 86f40f535d7c
Revises: ef29ba43e290
Create Date: 2022-05-10 16:55:28.541847

"""

# revision identifiers, used by Alembic.
revision = '86f40f535d7c'
down_revision = 'ef29ba43e290'

from alembic import op


def upgrade():
    index_data = [('challenge', 'timestamp'),
                  ('clientapplication', 'lastseen'),
                  ('pidea_audit', 'date'),
                  ('usercache', 'timestamp'),
                  ('authcache', 'first_auth'),
                  ('authcache', 'last_auth'),
                  ('monitoringstats', 'timestamp')]
    for k, v in index_data:
        try:
            op.create_index(op.f('ix_{0!s}_{1!s}'.format(k, v)), k, [v])
        except Exception as exx:
            print("Could not add index for column {1!s} in table {0!s}.".format(k, v))
            print(exx)


def downgrade():
    # there should be no need to disable/remove the index when downgrading.
    pass



import json
from privacyidea.models import (Resolver,
                                ResolverConfig,
                                db)
from .base import MyTestCase, MyApiTestCase


class ResolverModelTestCase(MyTestCase):
    '''
    Test the resolver on the database level
    '''
    
    def test_create_resolver(self):
        r = Resolver("r1", "passwdresolver")
        db.session.add(r)
        db.session.commit()
        self.assertTrue(r.name is not None, r.name)
        self.assertTrue(r.rtype is not None, r.rtype)
        # Add configuration to the resolver
        conf = ResolverConfig(r.id, "fileName", "somevalue")
        db.session.add(conf)
        db.session.commit()
        
        # Read Resolver
        r1 = Resolver.query.filter_by(name="r1").first()
        self.assertTrue(r1.rtype=="passwdresolver", r1.rtype)


class APIResolverTestCase(MyApiTestCase):
    '''
    Test the resolver on the API level
    '''
       
    def test_00_create_resolver(self):
        
        with self.app.test_request_context('/resolver/r1',
                                           data={"type": "passwdresolver",
                                                 "fileName": "/etc/passwd"},
                                           method='POST',
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertTrue(res.json['result']['status'], res.json)
            self.assertEqual(res.json['result']['value'], 1, res.json)
                    
        # check if the resolver was created
        with self.app.test_request_context('/resolver/',
                                           method='GET',
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertTrue(b'"fileName": "/etc/passwd"' in res.data, res.data)
            self.assertTrue(b'"resolvername": "r1"' in res.data, res.data)
            
    def test_01_create_realm(self):
        realm = "realm1"
        resolvers = "r1, r2"
        with self.app.test_request_context('/realm/{0!s}'.format(realm),
                                           data={"resolvers": resolvers},
                                           method='POST',
                                           headers={"Authorization": self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json
            value = result.get("result").get("value")
            self.assertTrue('r1' in value["added"], res.data)
            self.assertTrue('r2' in value["failed"], res.data)
            
'''
        with self.app.test_request_context(
                '/api/1.0/comments/' + str(c.id),
                method='PUT',
                data=json.dumps({'bad': 123}),
                headers={'Content-Type': 'application/json'}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401)
        '''



"""
This file contains the tests for lib/sqlutils.py
"""
from datetime import datetime

from mock import MagicMock
import warnings
from sqlalchemy.testing import AssertsCompiledSQL
from privacyidea.lib.sqlutils import DeleteLimit, delete_matching_rows
from privacyidea.models import Audit as LogEntry
from .base import MyTestCase


class SQLUtilsCompilationTestCase(MyTestCase, AssertsCompiledSQL):
    def test_01_delete_limit(self):
        stmt = DeleteLimit(LogEntry.__table__, LogEntry.id < 100)
        self.assertEqual(stmt.limit, 1000)
        stmt = DeleteLimit(LogEntry.__table__, LogEntry.id < 100, 1234)
        self.assertEqual(stmt.limit, 1234)
        with self.assertRaises(RuntimeError):
            DeleteLimit(LogEntry.__table__, LogEntry.id < 100, None)
        with self.assertRaises(RuntimeError):
            DeleteLimit(LogEntry.__table__, LogEntry.id < 100, "not an integer")
        with self.assertRaises(RuntimeError):
            DeleteLimit(LogEntry.__table__, LogEntry.id < 100, 0)
        with self.assertRaises(RuntimeError):
            DeleteLimit(LogEntry.__table__, LogEntry.id < 100, -10)

    def test_02_compile_delete_limit(self):
        now = datetime.now()
        stmt_age = DeleteLimit(LogEntry.__table__, LogEntry.date < now)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=BytesWarning)
            self.assert_compile(stmt_age,
                                "DELETE FROM pidea_audit WHERE pidea_audit.id IN "
                                "(SELECT pidea_audit.id FROM pidea_audit WHERE "
                                "pidea_audit.date < :date_1 LIMIT :param_1)",
                                checkparams={"date_1": now, "param_1": 1000},
                                dialect='default')
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=BytesWarning)
            self.assert_compile(stmt_age,
                                "DELETE FROM pidea_audit WHERE pidea_audit.date < %s LIMIT 1000",
                                checkpositional=(now,),
                                dialect='mysql')

        stmt_id = DeleteLimit(LogEntry.__table__, LogEntry.id < 1000, 1234)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=BytesWarning)
            self.assert_compile(stmt_id,
                                "DELETE FROM pidea_audit WHERE pidea_audit.id IN "
                                "(SELECT pidea_audit.id FROM pidea_audit WHERE "
                                "pidea_audit.id < :id_1 LIMIT :param_1)",
                                checkparams={"id_1": 1000, "param_1": 1234},
                                dialect='default')
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=BytesWarning)
            self.assert_compile(stmt_id,
                                "DELETE FROM pidea_audit WHERE pidea_audit.id < %s LIMIT 1234",
                                checkpositional=(1000,),
                                dialect='mysql')

    def test_03_delete(self):
        session = MagicMock()

        fake_results = [1000, 1000, 500]
        def fake_execute_delete1(stmt):
            result = MagicMock()
            result.rowcount = fake_results.pop(0)
            return result

        session.execute.side_effect = fake_execute_delete1
        # delete chunked
        result = delete_matching_rows(session, LogEntry.__table__, LogEntry.id < 1234, 1000)
        # was called three times to delete 2500 entries
        self.assertEqual(len(session.execute.mock_calls), 3)
        self.assertEqual(result, 2500)

        session.execute.reset_mock()
        def fake_execute_delete2(stmt):
            result = MagicMock()
            result.rowcount = 2500
            return result
        session.execute.side_effect = fake_execute_delete2
        # delete in one statement
        result = delete_matching_rows(session, LogEntry.__table__, LogEntry.id < 1234)
        self.assertEqual(len(session.execute.mock_calls), 1)
        self.assertEqual(result, 2500)


"""Enlarge serial column in audit table

Revision ID: 49a04e560d96
Revises: 2118e566df16
Create Date: 2018-02-14 16:20:20.609200

"""

# revision identifiers, used by Alembic.
revision = '49a04e560d96'
down_revision = '2118e566df16'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.alter_column('pidea_audit', 'serial',
               existing_type=sa.VARCHAR(length=20),
               type_=sa.String(length=40),
               existing_nullable=True)
    except Exception as exx:
        print("Could not enlarge column 'pidea_audit.serial'")
        print(exx)


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('pidea_audit', 'serial',
               existing_type=sa.String(length=40),
               type_=sa.VARCHAR(length=20),
               existing_nullable=True)
    ### end Alembic commands ###



"""v3.6: move PUSH registration URL from Firebase Config to policies

Revision ID: fa07bd604a75
Revises: 3ba618f6b820
Create Date: 2021-05-18 23:29:59.273457

"""

# revision identifiers, used by Alembic.
revision = 'fa07bd604a75'
down_revision = '3ba618f6b820'

from privacyidea.lib.smsprovider.SMSProvider import get_smsgateway, delete_smsgateway_option
from privacyidea.lib.tokens.pushtoken import GWTYPE, PUSH_ACTION
from privacyidea.lib.policy import PolicyClass, set_policy, SCOPE


def upgrade():
    # 1. Read the push_registration_url and ttl from the Firebase Config
    fb_gateways = get_smsgateway(gwtype=GWTYPE)
    print(fb_gateways)
    # 2. Check which policy contains this Firebase Config
    P = PolicyClass()
    pols = P.list_policies(scope=SCOPE.ENROLL,
                           action="{0!s}".format(PUSH_ACTION.FIREBASE_CONFIG))

    # iterate through all enrollment policies
    for pol in pols:
        # Check for all firebase gateways, if this policy needs to be modified
        for fbgw in fb_gateways:
            if pol.get("action").get(PUSH_ACTION.FIREBASE_CONFIG) == fbgw.identifier:
                print("Modifying policy {0!s}".format(pol.get("name")))
                # This is an enrollment policy, that references this very firebase config
                # 3. Add the push_registration_url and ttl to this policy
                registration_url = fbgw.option_dict.get("registration URL")
                ttl = fbgw.option_dict.get("time to live")
                # We can leave most of the parameters None, since it will update the policy.
                # We still need to pass the original "active" and "check_all_resolvers" params
                # and we need to update the action
                action = pol.get("action")
                # Only add registration_url and ttl to the policy, if these values actually exist,
                # to avoid deleting (setting an empty value) in the policy.
                if registration_url:
                    action[PUSH_ACTION.REGISTRATION_URL] = registration_url
                if ttl:
                    action[PUSH_ACTION.TTL] = ttl
                r = set_policy(name=pol.get("name"),
                               scope=SCOPE.ENROLL,
                               active=pol.get("active"),
                               check_all_resolvers=pol.get("check_all_resolvers"),
                               action=action)
                print("+- Updated policy {0!s}: {1!s}".format(pol.get("name"), r))
                # 4. Delete push_registration_url and ttl from the Firebase Config
                #    Note: If we had a firebase config, that would not be used in a policy,
                #    the url and ttl would not be deleted from the firebase config. But this
                #    does not matter. I like to keep it in this for-loop to avoid side unknown side effects.
                print("Deleting URL and TTL from the Firebase Gateway config.")
                if registration_url:
                    delete_smsgateway_option(fbgw.id, "registration URL")
                if ttl:
                    delete_smsgateway_option(fbgw.id, "time to live")


def downgrade():
    # The only way is up.
    pass



from .base import MyApiTestCase


class APIServiceIDTestCase(MyApiTestCase):

    def test_01_set_get_delete_serviceid(self):
        # Set serviceid
        with self.app.test_request_context('/serviceid/serviceA',
                                           data={"description": "My Cool first service"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertGreaterEqual(value, 1)

        # get services
        with self.app.test_request_context('/serviceid/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertIn("serviceA", value)
            self.assertEqual(value["serviceA"]["description"], "My Cool first service")

        # create a 2nd service
        with self.app.test_request_context('/serviceid/serviceB',
                                           data={"description": "2nd service"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertGreaterEqual(value, 1)

        # get all services
        with self.app.test_request_context('/serviceid/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertEqual(len(value), 2)
            self.assertIn("serviceA", value)
            self.assertIn("serviceB", value)

        # Change the description of the first group
        with self.app.test_request_context('/serviceid/serviceA',
                                           data={"description": "1st service"},
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertGreaterEqual(value, 1)

        # get 1st service
        with self.app.test_request_context('/serviceid/serviceA',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertEqual(len(value), 1)
            self.assertEqual(value["serviceA"]["description"], "1st service")

        # delete 1st service
        with self.app.test_request_context('/serviceid/serviceA',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertEqual(value, 1)

        # check, that only the 2nd service is available
        with self.app.test_request_context('/serviceid/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertEqual(len(value), 1)
            self.assertNotIn("serviceA", value)
            self.assertIn("serviceB", value)

        # delete 2nd service
        with self.app.test_request_context('/serviceid/serviceB',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            value = res.json['result']['value']
            self.assertEqual(value, 1)



#  2018-07-09 Friedrich Weber <friedrich.weber@netknights.it>
#             Initial implementation of periodic tasks
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """These endpoints are used to create, modify and delete
periodic tasks.

This module is tested in tests/test_api_periodictask.py"""

import json
import logging

from flask import Blueprint, g, request

from privacyidea.lib.config import get_privacyidea_node_names
from privacyidea.lib.tokenclass import AUTH_DATE_FORMAT
from privacyidea.api.lib.prepolicy import prepolicy, check_base_action
from privacyidea.api.lib.utils import send_result, getParam
from privacyidea.lib.error import ParameterError
from privacyidea.lib.policy import ACTION
from privacyidea.lib.log import log_with
from privacyidea.lib.periodictask import get_periodic_tasks, set_periodic_task, delete_periodic_task, \
    enable_periodic_task, get_periodic_task_by_id, get_taskmodule, get_available_taskmodules
from privacyidea.lib.utils import is_true

log = logging.getLogger(__name__)

periodictask_blueprint = Blueprint('periodictask_blueprint', __name__)


def convert_datetimes_to_string(ptask):
    """
    Convert the ``last_update`` and ``last_runs`` timestamps to ISO 8601 strings and return a copy of ``ptask``.

    :param ptask: periodic task dictionary
    :return: a new periodic task dictionary
    """
    ptask = ptask.copy()
    ptask['last_update'] = ptask['last_update'].strftime(AUTH_DATE_FORMAT)
    ptask['last_runs'] = dict((node, timestamp.strftime(AUTH_DATE_FORMAT))
                              for node, timestamp in ptask['last_runs'].items())
    return ptask


@periodictask_blueprint.route('/taskmodules/', methods=['GET'])
@log_with(log)
def list_taskmodules():
    """
    Return a list of task module identifiers.
    """
    taskmodules = get_available_taskmodules()
    g.audit_object.log({"success": True})
    return send_result(taskmodules)


@periodictask_blueprint.route('/nodes/', methods=['GET'])
@log_with(log)
def list_nodes():
    """
    Return a list of available nodes
    """
    nodes = get_privacyidea_node_names()
    g.audit_object.log({"success": True})
    return send_result(nodes)


@periodictask_blueprint.route('/options/<taskmodule>', methods=['GET'])
@log_with(log)
def get_taskmodule_options(taskmodule):
    """
    Return the available options for the given taskmodule.

    :param taskmodule: Identifier of the task module
    :return: a dictionary mapping option keys to description dictionaries
    """
    options = get_taskmodule(taskmodule).options
    g.audit_object.log({"success": True})
    return send_result(options)


@periodictask_blueprint.route('/', methods=['GET'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.PERIODICTASKREAD)
def list_periodic_tasks():
    """
    Return a list of objects of defined periodic tasks.
    """
    ptasks = get_periodic_tasks()
    result = [convert_datetimes_to_string(ptask) for ptask in ptasks]
    g.audit_object.log({"success": True})
    return send_result(result)


@periodictask_blueprint.route('/<ptaskid>', methods=['GET'])
@log_with(log)
@prepolicy(check_base_action, request, ACTION.PERIODICTASKREAD)
def get_periodic_task_api(ptaskid):
    """
    Return the dictionary describing a periodic task.

    :param ptaskid: ID of the periodic task
    """
    ptask = get_periodic_task_by_id(int(ptaskid))
    g.audit_object.log({"success": True})
    return send_result(convert_datetimes_to_string(ptask))


@periodictask_blueprint.route('/', methods=['POST'])
@prepolicy(check_base_action, request, ACTION.PERIODICTASKWRITE)
@log_with(log)
def set_periodic_task_api():
    """
    Create or replace an existing periodic task definition.

    :param id: ID of an existing periodic task definition that should be updated
    :param name: Name of the periodic task
    :param active: true if the periodic task should be active
    :param retry_if_failed: privacyIDEA will retry to execute the task if failed
    :param interval: Interval at which the periodic task should run (in cron syntax)
    :param nodes: Comma-separated list of nodes on which the periodic task should run
    :param taskmodule: Task module name of the task
    :param ordering: Ordering of the task, must be a number >= 0.
    :param options: A dictionary (possibly JSON) of periodic task options, mapping unicodes to unicodes
    :return: ID of the periodic task
    """
    param = request.all_data
    ptask_id = getParam(param, "id", optional=True)
    if ptask_id is not None:
        ptask_id = int(ptask_id)
    name = getParam(param, "name", optional=False)
    active = is_true(getParam(param, "active", default=True))
    retry_if_failed = is_true(getParam(param, "retry_if_failed", default=True))
    interval = getParam(param, "interval", optional=False)
    node_string = getParam(param, "nodes", optional=False)
    if node_string.strip():
        node_list = [node.strip() for node in node_string.split(",")]
    else:
        raise ParameterError("nodes: expected at least one node")
    taskmodule = getParam(param, "taskmodule", optional=False)
    if taskmodule not in get_available_taskmodules():
        raise ParameterError("Unknown task module: {!r}".format(taskmodule))
    ordering = int(getParam(param, "ordering", optional=False))
    options = getParam(param, "options", optional=True)
    if options is None:
        options = {}
    elif not isinstance(options, dict):
        options = json.loads(options)
        if not isinstance(options, dict):
            raise ParameterError("options: expected dictionary, got {!r}".format(options))
    result = set_periodic_task(name, interval, node_list, taskmodule, ordering, options, active, ptask_id,
                               retry_if_failed)
    g.audit_object.log({"success": True, "info": result})
    return send_result(result)


@periodictask_blueprint.route('/enable/<ptaskid>', methods=['POST'])
@prepolicy(check_base_action, request, ACTION.PERIODICTASKWRITE)
@log_with(log)
def enable_periodic_task_api(ptaskid):
    """
    Enable a certain periodic task.

    :param ptaskid: ID of the periodic task
    :return: ID of the periodic task
    """
    result = enable_periodic_task(int(ptaskid), True)
    g.audit_object.log({"success": True})
    return send_result(result)


@periodictask_blueprint.route('/disable/<ptaskid>', methods=['POST'])
@prepolicy(check_base_action, request, ACTION.PERIODICTASKWRITE)
@log_with(log)
def disable_periodic_task_api(ptaskid):
    """
    Disable a certain periodic task.

    :param ptaskid: ID of the periodic task
    :return: ID of the periodic task
    """
    result = enable_periodic_task(int(ptaskid), False)
    g.audit_object.log({"success": True})
    return send_result(result)


@periodictask_blueprint.route('/<ptaskid>', methods=['DELETE'])
@prepolicy(check_base_action, request, ACTION.PERIODICTASKWRITE)
@log_with(log)
def delete_periodic_task_api(ptaskid):
    """
    Delete a certain periodic task.

    :param ptaskid: ID of the periodic task
    :return: ID of the periodic task
    """
    result = delete_periodic_task(int(ptaskid))
    g.audit_object.log({"success": True, "info": result})
    return send_result(result)



# SPDX-FileCopyrightText: (C) 2024 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-FileCopyrightText: (C) 2018 Friedrich Weber <friedrich.weber@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
This script can be used to create a self-contained local privacyIDEA
instance that does not require a web server to run. Instead,
authentication requests are validated via the command line.

The ``create`` command launches a wizard that creates a new instance.
The ``configure`` command starts a local development server that can
be used to setup tokens. This server must not be exposed to the network!
The ``check`` command can then be used to authenticate users.
"""

import click
import functools
import json
import os
from pathlib import Path
import shutil
import string
import subprocess  # nosec B404 # only trusted input is used
from tempfile import NamedTemporaryFile

from privacyidea.app import create_app
from privacyidea.lib.security.default import DefaultSecurityModule
from privacyidea.lib.utils import get_version_number

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

# warnings.simplefilter("ignore", category=sqlalchemy.exc.SAWarning)

PI_CFG_TEMPLATE = """import os, logging

INSTANCE_DIRECTORY = os.path.abspath(os.path.dirname(__file__))
PI_ENCFILE = os.path.join(INSTANCE_DIRECTORY, 'encKey')
PI_AUDIT_KEY_PRIVATE = os.path.join(INSTANCE_DIRECTORY, 'private.pem')
PI_AUDIT_KEY_PUBLIC = os.path.join(INSTANCE_DIRECTORY, 'public.pem')
PI_AUDIT_SQL_TRUNCATE = True
PI_LOGFILE = os.path.join(INSTANCE_DIRECTORY, 'privacyidea.log')
PI_LOGLEVEL = logging.INFO
SQLALCHEMY_DATABASE_URI = 'sqlite:///' + os.path.join(INSTANCE_DIRECTORY, 'privacyidea.sqlite')

SECRET_KEY = b'{secret_key}'
PI_PEPPER = '{pi_pepper}'

"""

RSA_KEYSIZE = 2048
PEPPER_CHARSET = string.ascii_letters + string.digits + '_'


def invoke_pi_manage(commandline, pi_cfg):
    """
    Invoke ``pi-manage`` with arguments, setting PRIVACYIDEA_CONFIGFILE TO ``pi_cfg``.

    :param commandline: arguments to pass as a list
    :type commandline: list
    :param pi_cfg: location of the privacyIDEA config file
    :type pi_cfg: str or pathlib.Path
    """
    environment = os.environ.copy()
    environment['PRIVACYIDEA_CONFIGFILE'] = str(pi_cfg)
    subprocess.check_call(['pi-manage'] + commandline, env=environment)  # nosec B603 # only trusted input is used


def instance_option(f):
    """The instance option for all commands"""
    options = [
        click.option('-i', '--instance', default=Path.home() / '.privacyidea',
                     help='Location of the privacyIDEA instance', show_default=True,
                     type=click.Path(file_okay=False, resolve_path=True))
    ]
    return functools.reduce(lambda x, opt: opt(x), options, f)


@click.group(context_settings=CONTEXT_SETTINGS,
             epilog='Check out our docs at https://privacyidea.readthedocs.io/ for more details')
def cli():
    """
\b
             _                    _______  _______
   ___  ____(_)  _____ _______ __/  _/ _ \\/ __/ _ |
  / _ \\/ __/ / |/ / _ `/ __/ // // // // / _// __ |
 / .__/_/ /_/|___/\\_,_/\\__/\\_, /___/____/___/_/ |_|  Standalone
/_/                       /___/

  Management script for creating local privacyIDEA instances."""
    click.echo(r"""
             _                    _______  _______
   ___  ____(_)  _____ _______ __/  _/ _ \/ __/ _ |
  / _ \/ __/ / |/ / _ `/ __/ // // // // / _// __ |
 / .__/_/ /_/|___/\_,_/\__/\_, /___/____/___/_/ |_|  Standalone
/_/                       /___/
{0!s:>51}
    """.format('v{0!s}'.format(get_version_number())))


@click.command()
@instance_option
@click.pass_context
def configure(ctx, instance):
    """Run a local webserver to configure the privacyIDEA instance."""
    instance = Path(instance)
    if instance.exists() and instance.joinpath('pi.cfg').exists():
        app = create_app(config_name="production", config_file=instance.joinpath('pi.cfg'),
                         silent=True)
        app.run()
    else:
        click.secho(f"Instance configuration at '{instance}/pi.cfg' does "
                    f"not exist! Aborting.", fg='red')
        ctx.exit(1)


@click.command()
@instance_option
@click.pass_context
def create(ctx, instance):
    """ Create a new privacyIDEA instance."""
    instance = Path(instance)
    if instance.exists():
        click.secho(f"Instance at '{instance}' already exists! Aborting.", fg='red')
        ctx.exit(1)
    try:
        instance.mkdir()

        # create SECRET_KEY and PI_PEPPER
        secret_key = DefaultSecurityModule.random(24)
        pi_pepper = create_pepper()

        secret_key_hex = ''.join('\\x{:02x}'.format(b) for b in secret_key)
        # create a pi.cfg
        pi_cfg = instance / 'pi.cfg'
        with open(pi_cfg, 'w') as f:
            f.write(PI_CFG_TEMPLATE.format(
                secret_key=secret_key_hex,
                pi_pepper=pi_pepper
            ))

        # create an enckey
        invoke_pi_manage(['create_enckey'], pi_cfg)
        invoke_pi_manage(['create_audit_keys'], pi_cfg)
        invoke_pi_manage(['create_tables'], pi_cfg)

        click.secho('Please enter a password for the new admin `super`.', fg='blue')
        invoke_pi_manage(['admin', 'add', 'super'], pi_cfg)

        # create user resolver
        if click.confirm('Would you like to create a default resolver and realm?', default=True):
            click.echo("""
There are two possibilities to create a resolver:
 1) We can create a table in the privacyIDEA SQLite database to store the users.
    You can add users via the privacyIDEA Web UI.
 2) We can create a resolver that contains the users from /etc/passwd
    """)
            create_sql_resolver = click.prompt('Please choose (default=1): ',
                                               default=1, type=click.Choice(['1', '2']),
                                               show_choices=False, show_default=False)
            if create_sql_resolver == 1:
                invoke_pi_manage(['resolver', 'create_internal', 'defresolver'], pi_cfg)
            else:
                with NamedTemporaryFile(mode='w', delete=False) as f:
                    f.write('{"fileName": "/etc/passwd"}')
                invoke_pi_manage(['resolver', 'create', 'defresolver', 'passwdresolver',
                                  f.name], pi_cfg)
                os.unlink(f.name)
            invoke_pi_manage(['realm', 'create', 'defrealm', 'defresolver'], pi_cfg)

        click.secho('Configuration is complete. You can now configure privacyIDEA in '
                    'the web browser by running', fg='blue')
        click.secho(f"  privacyidea-standalone configure -i '{instance}' ", fg='blue')
    except Exception as _e:
        click.secho(f'Could not finish creation process! Removing '
                    f'instance directory {instance}.', fg='red')
        shutil.rmtree(instance)
        raise


@click.command()
@click.option('-r', '--response', 'show_response', is_flag=True,
              help='Print the JSON response of privacyIDEA to standard output')
@click.option("-u", "--username", prompt=True,
              help="The username to authenticate")
@click.option("-p", "--password", prompt=True, hide_input=True,
              help="The password to authenticate the user")
@instance_option
@click.pass_context
def check(ctx, instance, show_response, username, password):
    """
    Check the given username and password against privacyIDEA.
    This command reads two lines from standard input: The first line is
    the username, the second line is the password (which consists of a
    static part and the OTP).

    This commands exits with return code 0 if the user could be authenticated
    successfully.
    """
    exitcode = -1
    instance = Path(instance)
    if instance.exists() and instance.joinpath('pi.cfg').exists():
        try:
            app = create_app(config_name="production", config_file=instance.joinpath('pi.cfg'),
                             silent=True)
            with app.test_request_context('/validate/check', method='POST',
                                          data={'user': username, 'pass': password}):
                response = app.full_dispatch_request()
                if response.status_code == 200 and response.json['result']['value'] is True:
                    exitcode = 0
                else:
                    exitcode = 1
                if show_response:
                    click.secho(f"Status: {response.status}\n"
                                f"Data:\n{json.dumps(response.json, indent=2)}", fg='blue')
        except Exception as e:
            click.secho(str(e), fg='red')

    else:
        click.secho(f"Instance configuration at '{instance}/pi.cfg' does "
                    f"not exist! Aborting.", fg='red')

    ctx.exit(exitcode)


cli.add_command(configure)
cli.add_command(create)
cli.add_command(check)


def create_pepper(length=24, chunk_size=8, charset=PEPPER_CHARSET):
    """
    create a valid PI_PEPPER value of a given length from urandom,
    choosing characters from a given charset

    :param length: pepper length to generate
    :param chunk_size: number of bytes to read from urandom per iteration
    :param charset: list of valid characters
    :return: a string of the specified length
    :rtype: str
    """
    pepper = ''
    while len(pepper) < length:
        random_bytes = DefaultSecurityModule.random(chunk_size)
        printables = ''.join(chr(b) for b in random_bytes if chr(b) in charset)
        pepper += printables
    return pepper[:length]


if __name__ == '__main__':
    cli()



# 2015-06-05   Cornelius Kölbel <cornelius@privacyidea.org>
#              Add interface to edit and add users
# Dec 01, 2014 Cornelius Kölbel <cornelius@privacyidea.org>
#               Migration to flask
#               Adapt methods for tests
#               Improve comments
#               100% test code coverage
# 2014-10-03 fix getUsername function
#            Cornelius Kölbel <cornelius@privcyidea.org>
#  May, 08 2014 Cornelius Kölbel
#  http://www.privacyidea.org
#
#  product:  LinOTP2
#  module:   useridresolver
#  tool:     UserIdResolver
#  edition:  Comunity Edition
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module implements the communication interface
for resolving user info to the user base

Defines the rough interface for a UserId Resolver
== a UserId Resolver is required to resolve the
   Login Name to an unique User Identifier

- for /etc/passwd this will be the uid
- for ldap this might be the DN
- for SQL the unique index ( what's the right name here (tm))

"""


class UserIdResolver(object):

    fields = {"username": 1, "userid": 1,
              "description": 0,
              "phone": 0, "mobile": 0, "email": 0,
              "givenname": 0, "surname": 0, "gender": 0
              }
    name = ""
    id = "baseid"

    # If the resolver could be configured editable
    updateable = False

    def close(self):
        """
        Hook to close down the resolver after one request
        """
        return

    @staticmethod
    def getResolverClassType():
        """
        provide the resolver type for registration
        """
        return 'UserIdResolver'

    @staticmethod
    def getResolverType():
        """
        getResolverType - return the type of the resolver

        :return: returns the string 'ldapresolver'
        :rtype:  string
        """
        return 'UserIdResolver'

    @classmethod
    def getResolverClassDescriptor(cls):
        """
        return the descriptor of the resolver, which is
        - the class name and
        - the config description

        :return: resolver description dict
        :rtype:  dict
        """
        descriptor = {}
        typ = cls.getResolverClassType()
        descriptor['clazz'] = "useridresolver.UserIdResolver"
        descriptor['config'] = {}
        return {typ: descriptor}

    @staticmethod
    def getResolverDescriptor():
        """
        return the descriptor of the resolver, which is
        - the class name and
        - the config description

        :return: resolver description dict
        :rtype:  dict
        """
        return UserIdResolver.getResolverClassDescriptor()

    def getUserId(self, loginName):
        """
        The loginname is resolved to a user_id.
        Depending on the resolver type the user_id can
        be an ID (like in /etc/passwd) or a string (like
        the DN in LDAP)

        It needs to return an empty string, if the user does
        not exist.

        :param loginName: The login name of the user
        :type loginName: sting
        :return: The ID of the user
        :rtype: str
        """
        return "dummy_user_id"

    def getUsername(self, userid):
        """
        Returns the username/loginname for a given userid

        :param userid: The userid in this resolver
        :type userid: string
        :return: username
        :rtype: string
        """
        return "dummy_user_name"

    def getUserInfo(self, userid):
        """
        This function returns all user information for a given user object
        identified by UserID.

        :param userid: ID of the user in the resolver
        :type userid: int or string
        :return:  dictionary, if no object is found, the dictionary is empty
        :rtype: dict
        """
        return {}

    def getUserList(self, searchDict=None):
        """
        This function finds the user objects,
        that have the term 'value' in the user object field 'key'

        :param searchDict:  dict with key values of user attributes -
                    the key may be something like 'loginname' or 'email'
                    the value is a regular expression.
        :type searchDict: dict

        :return: list of dictionaries (each dictionary contains a
                 user object) or an empty string if no object is found.
        :rtype: list of dicts
        """
        searchDict = searchDict or {}
        return [{}]

    def getResolverId(self):
        """
        get resolver specific information
        :return: the resolver identifier string - empty string if not exist
        """
        return self.id

    def loadConfig(self, config):
        """
        Load the configuration from the dict into the Resolver object.
        If attributes are missing, need to set default values.
        If required attributes are missing, this should raise an
        Exception.

        :param config: The configuration values of the resolver
        :type config: dict
        """
        return self

    def checkPass(self, uid, password):
        """
        This function checks the password for a given uid.
        returns true in case of success
        false if password does not match

        :param uid: The uid in the resolver
        :type uid: string or int
        :param password: the password to check. Usually in cleartext
        :type password: string
        :return: True or False
        :rtype: bool
        """
        return False

    def add_user(self, attributes=None):
        """
        Add a new user in the useridresolver.
        This is only possible, if the UserIdResolver supports this and if
        we have write access to the user store.

        :param username: The login name of the user
        :type username: basestring
        :param attributes: Attributes according to the attribute mapping
        :return: The new UID of the user. The UserIdResolver needs to
        determine the way how to create the UID.
        """
        attributes = attributes or {}
        return None

    def delete_user(self, uid):
        """
        Delete a user from the useridresolver.
        The user is referenced by the user id.
        :param uid: The uid of the user object, that should be deleted.
        :type uid: basestring
        :return: Returns True in case of success
        :rtype: bool
        """
        return None

    def update_user(self, uid, attributes=None):
        """
        Update an existing user.
        This function is also used to update the password. Since the
        attribute mapping know, which field contains the password,
        this function can also take care for password changing.

        Attributes that are not contained in the dict attributes are not
        modified.

        :param uid: The uid of the user object in the resolver.
        :type uid: basestring
        :param attributes: Attributes to be updated.
        :type attributes: dict
        :return: True in case of success
        """
        attributes = attributes or {}
        return None

    @classmethod
    def testconnection(cls, param):
        """
        This function lets you test if the parameters can be used to create a
        working resolver.
        The implementation should try to connect to the user store and verify
        if users can be retrieved.
        In case of success it should return a text like
        "Resolver config seems OK. 123 Users found."

        :param param: The parameters that should be saved as the resolver
        :type param: dict
        :return: returns True in case of success and a descriptive text
        :rtype: tuple
        """
        success = False
        desc = "Not implemented"
        return success, desc

    @property
    def editable(self):
        """
        Return true, if the Instance! of this resolver is configured editable.
        :return:
        """
        return False

    @property
    def has_multiple_loginnames(self):
        """
        Return if this resolver has multiple loginname attributes
        :return: bool
        """
        return False
